# -*- coding: utf-8 -*-

require 'date'
require 'forwardable'

class BitStream
  MASK = 0.upto(32).map do |n|
    (2 ** n) - 1
  end
  def initialize(s)
    @buf = s
    @ptr = 0
  end
  def get1
    getn(1).unpack("C")[0]
  end
  def get2
    getn(2).unpack("n")[0]
  end
  def get3
    getn(2).unpack("n")[0] * 256 + getn(1).unpack("C")[0]
  end
  def get4
    getn(4).unpack("N")[0]
  end
  def getn(n)
    rslt = @buf[@ptr, n]
    @ptr += n
    rslt
  end
  def getfield(field)
    case field
    when 1
      get1
    when 2
      get2
    when 3
      get3
    when 4
      get4
    else
      getn(field)
    end
  end
  def getall
    rslt = @buf[@ptr..-1]
    @ptr = @buf.length
    rslt
  end
  def get1bit(*field)
    getbit(get1, field)
  end
  def get2bit(*field)
    getbit(get2, field)
  end
  def get3bit(*field)
    getbit(get3, field)
  end
  def get4bit(*field)
    getbit(get4, field)
  end
  def getary(n, *field)
    if block_given?
      Array.new(n) { yield }
    else
      if field.size > 1
        Array.new(n) { field.map { |f| getfield(f) } }
      else
        Array.new(n) { getfield(field[0]) }
      end
    end
  end
  def getallary(*field)
    if block_given?
      map { yield }
    else
      if field.size > 1
        map { field.map { |f| getfield(f) } }
      else
        map { getfield(field[0]) }
      end
    end
  end
  def eos
    (@buf.length <= @ptr)
  end
  def length
    @buf.length
  end
  def ptr
    @ptr
  end

  include Enumerable
  def each
    until (@buf.length <= @ptr)
      yield
    end
  end

  def self.bcd2int(bcd, bitlen)
    rslt = 0
    n = 1
    while bitlen > 0
      rslt += (bcd & 0xf) * n
      n *= 10
      bcd >>= 4
      bitlen -= 4
    end
    rslt
  end
  def getbcd(field)
    BitStream.bcd2int(getfield(field), field * 8)
  end
  def get1bcd
    n = get1
    n / 16 * 10 +
      n % 16
  end
  def get2bcd
    n = get2
    n / 16 ** 3 * 10 ** 3 +
      n / 16 ** 2 % 16 * 10 ** 2 +
      n / 16 % 16 * 10 +
      n % 16
  end
  def get3bcd
    n = get3
    n / 16 ** 5 * 10 ** 5 +
      n / 16 ** 4 % 16 * 10 ** 4 +
      n / 16 ** 3 % 16 * 10 ** 3 +
      n / 16 ** 2 % 16 * 10 ** 2 +
      n / 16 % 16 * 10 +
      n % 16
  end
  def get4bcd
    n = get4
    n / 16 ** 7 * 10 ** 7 +
      n / 16 ** 6 % 16 * 10 ** 6 +
      n / 16 ** 5 % 16 * 10 ** 5 +
      n / 16 ** 4 % 16 * 10 ** 4 +
      n / 16 ** 3 % 16 * 10 ** 3 +
      n / 16 ** 2 % 16 * 10 ** 2 +
      n / 16 % 16 * 10 +
      n % 16
  end
  def getmjd
    Date.jd(get2 + Date::MJD_EPOCH_IN_CJD)
  end
  Jtc = Struct.new(:hour, :min, :sec)
  def getjtc
    t = get3bcd
    Jtc.new(t / 10000, t / 100 % 100, t % 100)
  end
  def getmjdjtc
    d = get2
    t = getjtc
    DateTime.jd(d + Date::MJD_EPOCH_IN_CJD,
                t.hour, t.min, t.sec)
  end

  private
  def getbit(work, field)
    rslt = []
    field.reverse_each do |n|
      if n < 0
        rslt.unshift(BitStream.bcd2int(work & MASK[-n], -n))
        work >>= -n
      else
        rslt.unshift(work & MASK[n])
        work >>= n
      end
    end
    rslt
  end
end

module Descriptor
  TAG_TABLE = {}
  def Descriptor.def_tag(klass, tag)
    TAG_TABLE[tag] = klass
  end
  class Common
    def initialize(table, descriptor_tag, descriptor_length, body)
      @descriptor_tag = descriptor_tag
      @descriptor_length = descriptor_length
    end
    attr_reader :table, :descriptor_tag
    # 小クラス用(繰り返し部分を@dataとする)
    attr_reader :data
    def each
      @data.each do |data|
        yield *data
      end
    end
  end
  # nameのみの記述子
  class Name < Common
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      @name = table.text(body)
    end
    attr_reader :name
  end
  # 限定受信方式記述子
  class ConditionalAccess < Common
    Descriptor::def_tag self, 0x09
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @ca_system_id = bs.get2
      @ca_pid = bs.get2bit(3, 13)[1]
      @private_data_byte = bs.getall
    end
  end
  # ネットワーク名記述子
  class NetworkName < Name
    Descriptor::def_tag self, 0x40
  end
  # サービスリスト記述子
  class ServiceList < Common
    Descriptor::def_tag self, 0x41
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @data = bs.getallary(2,   # service_id サービス識別
                           1)   # service_type サービス形式種別
    end
  end
  # スタッフ記述子
  class Stuffing < Common
    Descriptor::def_tag self, 0x42
  end
  # 衛生分配システム記述子
  class SatelliteDeliverySystem < Common
    Descriptor::def_tag self, 0x43
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      #@frequency = BitStream.bcd2int(bs.get4, 32)
      @frequency = bs.get4bcd
      #@orbital_position = BitStream.bcd2int(bs.get2, 16)
      @orbital_position = bs.get2bcd
      @west_east_flag, @polarisation, @modulation = bs.get1bit(1, 2, 5)
      symbol_rate, @fec_inner = bs.get4bit(28, 4)
      @symbol_rate = BitStream.bcd2int(symbol_rate, 28)
    end
    attr_reader :frequency, :symbol_rate
  end
  # ブーケ名記述子
  class BouquetName < Name
    Descriptor::def_tag self, 0x47
  end
  # サービス記述子
  class Service < Common
    Descriptor::def_tag self, 0x48
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @service_type = bs.get1  # サービス形式種別
      service_provider_name_length = bs.get1
      @service_provider_name = table.text(bs.getn(service_provider_name_length))
      service_name_length = bs.get1
      @service_name = table.text(bs.getn(service_name_length))
    end
    attr_reader :service_type, :service_provider_name, :service_name
  end
  # 短形式イベント記述子
  class ShortEvent < Common
    Descriptor::def_tag self, 0x4d
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @iso_639_language_code = bs.getn(3)
      event_name_length = bs.get1
      @event_name = table.text(bs.getn(event_name_length))
      text_length = bs.get1
      @text = table.text(bs.getn(text_length))
    end
    attr_reader :iso_639_language_code, :event_name, :text
  end
  # 拡張形式イベント記述子
  class ExtendedEvent < Common
    Descriptor::def_tag self, 0x4e
    Data = Struct.new(:item_description, :item)
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @descriptor_number, @last_descriptor_number = bs.get1bit(4, 4)
      @iso_639_language_code = bs.getn(3)
      length_of_items = bs.get1
      data = BitStream.new(bs.getn(length_of_items))
      @data = data.getallary do
        item_description_length = data.get1
        item_description = table.text(data.getn(item_description_length))
        item_length = data.get1
        item = data.getn(item_length)
        Data.new(item_description, item)
      end
      text_length = bs.get1
      @text = table.text(bs.getn(text_length))
    end
    attr_reader :descriptor_number, :last_descriptor_number
    attr_reader :iso_639_language_code, :text
  end
  # コンポーネント記述子
  class Component < Name
    Descriptor::def_tag self, 0x50
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @stream_content = bs.get1bit(4, 4)[1]
      @componet_type = bs.get1
      @componet_tag = bs.get1
      @iso_639_language_code = bs.getn(3)
      @text = table.text(bs.getall)
    end
    attr_reader :stream_content, :componet_type, :component_tag,
    :iso_639_language_code, :text
  end
  # ストリーム識別記述子
  class StreamIdentifier < Name
    Descriptor::def_tag self, 0x52
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      @component_tag = body.ord
    end
    attr_reader :component_tag
  end
  # コンテント記述子
  class Content < Common
    Descriptor::def_tag self, 0x54
    Data = Struct.new(:content_nibble_level_1, :content_nibble_level_2,
                      :user_nibble1, :user_nibble2)
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @data = bs.getallary do
        Data.new(*bs.get2bit(4, 4, 4, 4))
      end
    end
    attr_reader :component_tag
  end
  # デジタルコピー制御記述子
  class DigitalCopyControl < Common
    Descriptor::def_tag self, 0xc1
    # 省略
  end
  # 音声コンポーネント記述子
  class AudioComponent < Name
    Descriptor::def_tag self, 0xc4
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @stream_content = bs.get1bit(4, 4)[1]
      @componet_type = bs.get1
      @componet_tag = bs.get1
      @stream_type = bs.get1
      @simulcast_group_tag = bs.get1
      @es_multi_lingual_flag, @main_component_flag,
      @quality_indicator, @sampling_rate, = bs.get1bit(1, 1, 2, 3, 1)
      @iso_639_language_code = bs.getn(3)
      if @es_multi_lingual_flag == 1
        @iso_639_language_code_2 = bs.getn(3)
      end
      @text = table.text(bs.getall)
    end
    attr_reader :stream_content, :componet_type, :component_tag,
    :stream_type, :simulcast_group_tag, :es_multi_lingual_flag,
    :main_component_flag, :quality_indicator, :sampling_rate,
    :iso_639_language_code, :iso_639_language_code_2, :text
  end
  # データコンテンツ記述子
  class VideoDecodeControl < Common
    Descriptor::def_tag self, 0xc7
  end
  # ビデオコントロール記述子
  class VideoDecodeControl < Common
    Descriptor::def_tag self, 0xc8
  end
  # CA契約情報記述子
  class CAContractInfo < Common
    Descriptor::def_tag self, 0xcb
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @ca_system_id = bs.get2
      @ca_unit_id, num_of_component = bs.get1bit(4, 4)
      @component_tag = bs.getn(num_of_component)
      contract_verification_info_length = bs.get1
      @contract_verification_info = bs.getn(contract_verification_info_length)
      fee_name_length = bs.get1
      @fee_name = bs.getn(fee_name_length)
    end
    attr_reader :ca_system_id, :ca_unit_id, :componet_tag
    attr_reader :contact_verification_info, :fee_name
  end
  # CAサービス記述子
  class CAService < Common
    Descriptor::def_tag self, 0xcc
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @ca_system_id = bs.get2
      @ca_broadcaster_group_id = bs.get1
      @message_control = bs.get1
      @service_id = bs.getallary(2)
    end
    attr_reader :ca_system_id, :ca_broadcaster_group_id
    attr_reader :message_control, :service_id
  end
  # TS情報記述子
  class TSInformation < Common
    Descriptor::def_tag self, 0xcd
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @remote_control_key_id = bs.get1
      length_of_ts_name = bs.get1bit(6, 2)[0]
      @ts_name = table.text(bs.getn(length_of_ts_name))
    end
    attr_reader :remote_control_key_id, :ts_name
  end
  # 拡張ブロードキャスタ記述子
  class ExtendedBroadcaster < Common
    Descriptor::def_tag self, 0xc3
    Broadcaster = Struct.new(:original_network_id, :broadcaster_id)
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @broadcaster_type = bs.get1bit(4, 4)[0]
      if @broadcaster_type == 0x1
        terrestrial_broadcaster_id = bs.get2
        number_of_affiliation_id_loop,
        number_of_broadcaster_id_loop = bs.get(4, 4)
        @affiliation_id = bs.getary(number_of_affiliation_id_loop, 1)
        @broadcaster_id = bs.getary(number_of_broadcaster_id_loop) do
          Broadcaster.new(bs.get2, bs.get1)
        end
      end
    end
    attr_reader :broadcaster_type, :affiliation_id, :broadcaster_id
  end
  # ロゴ伝送記述子
  class LogoTransmission < Common
    Descriptor::def_tag self, 0xcf
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      logo_transmission_type = bs.get1
      case logo_transmission_type
      when 0x01
        @logo = {
          :logo_transmission_type => logo_transmission_type,
          :log_id => bs.get2bit(7, 9)[1],
          :logo_version => bs.get2bit(4, 12)[1],
          :download_data_id => bs.get2,
        }
      when 0x02
        @logo = {
          :logo_transmission_type => logo_transmission_type,
          :log_id => bs.get2bit(7, 9)[1]
        }
      when 0x03
        @logo = {
          :logo_transmission_type => logo_transmission_type,
          :logo_char => table.text(bs.getall)
        }
      when
        @logo = nil
      end
    end
    attr_reader :logo
  end
  # シリーズ記述子
  class Series < Common
    Descriptor::def_tag self, 0xd5
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @series_id = bs.get2
      @repeat_label, @program_pattern, @expire_date_valid_flag =
        bs.get1bit(4, 3, 1)
      @expire_date = bs.getmjd
      @episode_number, @last_episode_number = bs.get3bit(12, 12)
      @series_name = table.text(bs.getall)
    end
    attr_reader :series_id, :repeat_label, :program_pattern
    attr_reader :expire_date_valid_flag, :expire_date
    attr_reader :episode_number, :last_episode_number, :series_name
  end
  # イベントグループ記述子
  class EventGroup < Common
    Descriptor::def_tag self, 0xd6
    Data = Struct.new(:service_id, :event_id)
    OtherData = Struct.new(:original_network_id, :transport_stream_id,
                           :service_id, :event_id)
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @group_type, event_count = bs.get1bit(4, 4)
      @data = bs.getary(event_count) do
        Data.new(bs.get2, bs.get2)
      end
      if @group_type == 4 || @group_type == 5
        @otherdata = bs.getallary do
          OtherData.new(bs.get2, bs.get2, bs.get2, bs.get2)
        end
      end
    end
    attr_reader :group_type, :otherdata
  end
  class SI < Common
    Table = Struct.new(:nit, :nit_o, :bit, :sdt, :sdt_o, :eitpf, :eitpf_o,
                       :eitsb, :eitse, :eitsb_o, :sdtt, :cdt,
                       :nbitm, :nbitr, :ldt)
    Eitpf = Struct.new(:table_cycle_h, :table_cycle_m, :table_cycle_l,
                       :num_of_meit_event, :num_of_leit_event)
    Media = Struct.new(:tv, :au, :dt)
    Eits = Struct.new(:media_type, :pattern, :schedule_range,
                      :base_cycle, :cycle_group)
    CycleGroup = Struct.new(:num_of_segment, :cycle)
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
    end
    def eitpf(desc)
      if desc.length == 1
        # BS, CS110
        desc.get1bcd
      else
        # TE
        Eitpf.new(desc.get1bcd,
                  desc.get1bcd,
                  desc.get1bcd,
                  *desc.get1bit(4, 4))
      end
    end
    def eits(desc)
      media = Media.new(Eits.new, Eits.new, Eits.new)
      media.tv.cycle_group = [ CycleGroup.new ]
      media.au.cycle_group = [ CycleGroup.new ]
      media.dt.cycle_group = [ CycleGroup.new ]
      desc.getallary do
        media_type, pattern = *desc.get1bit(2, 2, 4)[0, 2]
        data = case media_type
               when 1
                 media.tv
               when 2
                 media.au
               when 3
                 media.dt
               end
        data.media_type, data.pattern = media_type, pattern
        data.schedule_range = desc.get1bcd
        data.base_cycle, reserved,
        cycle_group_count = desc.get2bit(-12, 2, 2)
        if cycle_group_count == 0
          data.cycle_group = [ CycleGroup.new ]
        else
          data.cycle_group = desc.getary(cycle_group_count) do
            CycleGroup.new(desc.get1bcd, desc.get1bcd)
          end
        end
      end
      media
    end
  end
  # SI伝送パラメータ記述子
  class SIParameter < SI
    Descriptor::def_tag self, 0xd7
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @parameter_version = bs.get1
      @update_time = bs.getmjd
      @table = Table.new
      until bs.eos
        table_id = bs.get1
        table_description_length = bs.get1
        desc = BitStream.new(bs.getn(table_description_length))
        case table_id
        when 0x40               # TE1, BS1, CS1
          @table.nit = desc.get1bcd
        when 0x41               # CS1
          @table.nit_o = desc.get1bcd
        when 0xc4               # TE1, BS1, CS1
          @table.bit = desc.get1bcd
        when 0x42               # TE1, BS1, CS1
          @table.sdt = desc.get1bcd
        when 0x46               # BS1, CS1
          @table.sdt_o = desc.get1bcd
        when 0x4e               # TE1, TE2, BS1, CS1
          @table.eitpf = eitpf(desc)
        when 0x4f               # BS1, CS1
          @table.eitpf_o = eitpf(desc)
        when 0x50               # TE1, TE2, BS1, CS1, CS2
          @table.eitsb = eits(desc)
        when 0x58               # TE2, CS2
          @table.eitse = eits(desc)
        when 0x60               # BS1, CS1
          @table.eitsb_o = eits(desc)
        when 0xc3               # TE2, BS1, CS1
          @table.sdtt = desc.get2bcd
        when 0xc6               # CS1
          @table.nbitr = desc.get1bcd
        when 0xc7               # CS2
          @table.ldt = desc.get2bcd
        when 0xc8               # TE2
          @table.cdt = desc.get2bcd
        end
      end
    end
    attr_reader :parameter_version, :update_time, :table
  end
  # ブロードキャスタ名記述子
  class BroadcasterName < Name
    Descriptor::def_tag self, 0xd8
  end
  # コンポーネントグループ記述子
  class ComponentGroup < Common
    Descriptor::def_tag self, 0xd9
  end
  # SIプライムTS記述子
  class SIPrimeTS < SI
    Descriptor::def_tag self, 0xda
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @parameter_version = bs.get1
      @update_time = bs.getmjd
      @si_prime_ts_network_id = bs.get2
      @si_prime_transport_stream_id = bs.get2
      @table = Table.new
      until bs.eos
        table_id = bs.get1
        table_description_length = bs.get1
        desc = BitStream.new(bs.getn(table_description_length))
        case table_id
        when 0x42
          @table.sdt = desc.get1bcd
        when 0x46
          @table.sdt_o = desc.get1bcd
        when 0x4e
          @table.eitpf = eitpf(desc)
        when 0x4f
          @table.eitpf_o = eitpf(desc)
        when 0x50
          @table.eitsb = eits(desc)
        when 0x60
          @table.eitsb_o = eits(desc)
        when 0xc5
          @table.nbitm = desc.get1bcd
        when 0xc6
          @table.nbitr = desc.get1bcd
        end
      end
    end
    attr_reader :parameter_version, :update_time, :table
    attr_reader :si_prime_ts_network_id, :si_prime_transport_stream_id

  end
  # 地上分配システム記述子
  class TerrestrialDeliverySystem < Common
    Descriptor::def_tag self, 0xfa
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @area_code, @guard_interval, @transmission_mode =
        bs.get2bit(12, 2, 2)
      @frequency = bs.getallary(2)
    end
    attr_reader :area_code, :guard_interval, :transmission_mode, :frequency
  end
  # 緊急情報記述子
  class EmergencyInformation < Common
    Descriptor::def_tag self, 0xfc
  end
  # データ符号化方式記述子
  class DataComponent < Common
    Descriptor::def_tag self, 0xfd
    def initialize(table, descriptor_tag, descriptor_length, body)
      super
      bs = BitStream.new(body)
      @data_component_id = bs.get2
      @additional_data_component_info = bs.getall
    end
    attr_reader :data_component_id, :additional_data_component_info
  end
  # システム管理記述子
  class SystemManagement < Common
    Descriptor::def_tag self, 0xfe
  end
  def Descriptor.analyze(table, bs)
    descriptor_tag = bs.get1
    descriptor_length = bs.get1
    body = bs.getn(descriptor_length)

    klass = TAG_TABLE[descriptor_tag]
    klass = Common unless klass
    #printf "%02x %s\n", descriptor_tag, klass.inspect
    klass.new(table, descriptor_tag, descriptor_length, body)
  end
end

module PID
  # Table.def_tagで小細工
end

module Table
  TAG_TABLE = {}
  TIMEOUT_TABLE = {}
  def Table.def_tag(klass, pid, tags, timeouts = nil)
    name = klass.name.split('::').last
    PID.module_eval %Q{
      const_set("#{name}", pid)
    }
    klass.module_eval %Q{
      const_set(:TAG, tags[0])
      const_set(:TAGs, tags)
    }
    tags.each_with_index do |tag, idx|
      if tag.class == Range
        tag.each do |n|
          TAG_TABLE[n] = klass
        end
      else
        TAG_TABLE[tag] = klass
      end
      if timeouts
        TIMEOUT_TABLE[tag] = {}
        timeouts.keys.each do |type|
          TIMEOUT_TABLE[tag][type] = timeouts[type][idx]
        end
      end
    end
  end
  def Table.timeout(type, *tags)
    min = nil
    tags.each do |tag|
      next unless TIMEOUT_TABLE[tag][type]
      if min == nil || TIMEOUT_TABLE[tag][type] < min
        min = TIMEOUT_TABLE[tag][type]
      end
    end
    min = 3000 if min && min < 3000
    min
  end

  class Common
    extend Forwardable
    def_delegators(:@psisi, :table_id)
    def_delegators(:@psisi, :section_syntax_indicator)
    def_delegators(:@psisi, :section_length)
    def_delegators(:@psisi, :table_id_extention)
    def_delegators(:@psisi, :version_number)
    def_delegators(:@psisi, :current_next_indicator)
    def_delegators(:@psisi, :section_number)
    def_delegators(:@psisi, :last_section_number)
    def_delegators(:@psisi, :pid)
    def_delegators(:@psisi, :text) # 8単位符号系変換
    def_delegators(:@psisi, :body) # last_section_number の次からcrc32手前まで
    def_delegators(:@psisi, :PAT)
    def_delegators(:@psisi, :pat_data)
    def_delegators(:@psisi, :PMT)
    def_delegators(:@psisi, :pmt_pcr_pid)
    def_delegators(:@psisi, :pmt_program_info)
    def_delegators(:@psisi, :pmt_data)
    def initialize(psisi)
      @psisi = psisi
    end
    # 小クラス用(繰り返し部分を@dataとする)
    attr_reader :data
    def each
      @data.each do |data|
        yield data
      end
    end
  end

  # vvvvv PSI vvvvv
  # Program association Table
  class PAT < Common
    Table::def_tag self, 0x0000, [Psisi::PAT],
    { :te=>[100], :bs=>[100], :cs110=>[100] }
    Data = Struct.new(:program_number, # サービス識別とイコール
                      :pid)
    def initialize(psisi)
      super
      if false
        bs = BitStream.new(body)
        @data = bs.getallary do
          Data.new(bs.get2, bs.get2bit(3, 13)[1])
        end
      else
        @data = pat_data.map { |v| Data.new(*v) }
      end
    end
    alias :transport_stream_id :table_id_extention
  end # PAT
  # Conditional Access Table
  class CAT < Common
    Table::def_tag self, 0x0001, [0x01],
    { :te=>[10000], :bs=>[10000], :cs110=>[10000] }
    def initialize(psisi)
      super
      desc = BitStream.new(body)
      @descriptor = desc.getallary do
        Descriptor.analyze(self, desc)
      end
    end
    attr_reader :descriptor
  end # CAT
  # Program Map Table
  class PMT < Common
    Table::def_tag self, 0xffff, [Psisi::PMT],
    { :te=>[200], :bs=>[100], :cs110=>[100] }
    Data = Struct.new(:stream_type, :elementary_pid, :descriptor)
    def initialize(psisi)
      super
      if false
        bs = BitStream.new(body)
        @pcr_pid = bs.get2bit(3, 13)[1]

        program_info_length = bs.get2bit(4, 12)[1]
        desc = BitStream.new(bs.getn(program_info_length))
        @descriptor = desc.getallary do
          Descriptor.analyze(self, desc)
        end

        @data = bs.getallary do
          data = Data.new(bs.get1, bs.get2bit(3, 13)[1])
          es_info_length = bs.get2bit(4, 12)[1]
          desc = BitStream.new(bs.getn(es_info_length))
          data.descriptor = desc.getallary do
            Descriptor.analyze(self, desc)
          end
          data
        end
      else
        @pcr_pid = pmt_pcr_pid

        desc = BitStream.new(pmt_program_info)
        @descriptor = desc.getallary do
          Descriptor.analyze(self, desc)
        end

        @data = pmt_data.map do |stream_type, elementary_pid, es_info|
          data = Data.new(stream_type, elementary_pid)
          desc = BitStream.new(es_info)
          data.descriptor = desc.getallary do
            Descriptor.analyze(self, desc)
          end
          data
        end
      end
    end
    alias :program_number :table_id_extention # サービス識別とイコール
    attr_reader :descriptor
  end
  # NIT, BAT 共通
  class NITBAT < Common
    Data = Struct.new(:transport_stream_id,
                      :original_network_id, # ネットワーク識別
                      :descriptor)
    def initialize(psisi)
      super
      bs = BitStream.new(body)

      network_descriptors_length = bs.get2bit(4, 12)[1]
      desc = BitStream.new(bs.getn(network_descriptors_length))
      @descriptor = desc.getallary do
        Descriptor.analyze(self, desc)
      end

      transport_stream_loop_length = bs.get2bit(4, 12)[1]
      @data = bs.getallary do
        data = Data.new(bs.get2, bs.get2)
        transport_descriptors_length = bs.get2bit(4, 12)[1]
        desc = BitStream.new(bs.getn(transport_descriptors_length))
        data.descriptor = desc.getallary do
          Descriptor.analyze(self, desc)
        end
        data
      end
    end
    attr_reader :descriptor
  end # NITBAT
  # ネットワーク情報テーブル
  class NIT < NITBAT
    Table::def_tag self, 0x0010, [0x40, 0x41],
    { :te=>[3000, nil], :bs=>[20000, nil], :cs110=>[5000, 20000] }
    alias :network_id :table_id_extention # ネットワーク識別
  end # NIT
  # ^^^^^ PSI ^^^^^

  # vvvvv SI vvvvv
  # サービス記述テーブル
  class SDT < Common
    Table::def_tag self, 0x0011, [0x42, 0x46],
    { :te=>[3000, nil], :bs=>[5000, 20000], :cs110=>[5000, 20000] }
    Data = Struct.new(:service_id, # サービス識別
                      :heit_flag, :meit_flag, :leit_flag,
                      :eit_schedule_flag, :eit_present_following_flag,
                      :running_status, :free_ca_mode,
                      :descriptor)
    def initialize(psisi)
      super
      bs = BitStream.new(body[3..-1])
      @data = bs.getallary do
        data = Data.new(bs.get2, *bs.get1bit(3, 1, 1, 1, 1, 1)[1..-1])
        data.running_status, data.free_ca_mode,
        descriptors_loop_length = bs.get2bit(3, 1, 12)
        desc = BitStream.new(bs.getn(descriptors_loop_length))
        data.descriptor = desc.getallary do
          Descriptor.analyze(self, desc)
        end
        data
      end
    end
    alias :transport_stream_id :table_id_extention
    def original_network_id
      body[0, 2].unpack("n")[0]
    end
  end # SDT
  # イベント情報テーブル
  class EIT < Common
    Table::def_tag self, 0x0012, [0x4e,
                                  0x4f,
                                  0x50..0x57,
                                  0x58..0x5f,
                                  0x60..0x67,
                                  0x68..0x6f],
    { :te=>[nil], :bs=>[nil], :cs110=>[nil] }
    Data = Struct.new(:event_id, :start_time, :duration,
                      :running_status, :free_ca_mode, :descriptor)
    def type
      case table_id
      when 0x4e
        :eitpf
      when 0x4f
        :eitpf_o
      when 0x50..0x57
        :eitsb
      when 0x58..0x5f
        :eitse
      when 0x60..0x67
        :eitsb_o
      when 0x68..0x6f
        :eitse_o
      end
    end
    def initialize(psisi)
      super
      #printf "%02x: ", table_id
      bs = BitStream.new(body)
      @transport_stream_id = bs.get2
      @original_network_id = bs.get2
      @segment_last_section_number = bs.get1
      @last_table_id = bs.get1
      @data = bs.getallary do
        data = Data.new(bs.get2, bs.getmjdjtc, bs.getjtc)
        data.running_status, data.free_ca_mode,
        descriptors_loop_length = bs.get2bit(3, 1, 12)
        desc = BitStream.new(bs.getn(descriptors_loop_length))
        data.descriptor = desc.getallary do
          Descriptor.analyze(self, desc)
        end
        data
      end
    end
    alias :service_id :table_id_extention
    attr_reader :transport_stream_id, :original_network_id
    attr_reader :segment_last_section_number, :last_table_id
  end # EIT
  # ブロードキャスタ情報テーブル
  class BIT < Common
    Table::def_tag self, 0x0024, [0xc4],
    { :te=>[3000], :bs=>[20000], :cs110=>[20000] }
    Data = Struct.new(:broadcaster_id, :descriptor)
    def initialize(psisi)
      super
      bs = BitStream.new(body)

      first_descriptors_length = bs.get2bit(3, 1, 12)[2]
      desc = BitStream.new(bs.getn(first_descriptors_length))
      @descriptor = desc.getallary do
        Descriptor.analyze(self, desc)
      end

      @data = bs.getallary do
        data = Data.new(bs.get1)
        broadcaster_descriptors_length = bs.get2bit(4, 12)[1]
        desc = BitStream.new(bs.getn(broadcaster_descriptors_length))
        data.descriptor = desc.getallary do
          Descriptor.analyze(self, desc)
        end
        data
      end
    end
    alias :original_network_id :table_id_extention # ネットワーク識別
    attr_reader :descriptor
  end # BIT
  # ^^^^^ SI ^^^^^

  def Table.analyze(psisi)
    klass = TAG_TABLE[psisi.table_id]
    klass = Common unless klass
    klass.new(psisi)
  end
end

class CycleParameter
  All_Group2Range = {
    :eitsb_tv_b => :d1,         # TE, BS, CS110
    :eitsb_tv_e1 => :s1,        # TE, BS, CS110
    :eitsb_tv_e2 => :s2,        # TE, BS, CS110
    :eitsb_au_b => :d2,         # BS, CS110
    :eitsb_au_e => :s3,         # BS, CS110
    :eitsb_dt_b => :d3,         # TE, BS, CS110
    :eitsb_dt_e => :s4,         # TE, BS, CS110
  }
  Each_Group2Range = {
    :eitsb_tv_b => :d4,         # TE, BS
    :eitsb_au_b => :d5,         # BS
    :eitsb_dt_b => :d6,         # TE, BS
    :eitse_tv_b => :d7,         # TE, BS
    :eitse_tv_e => :s5,         # TE, BS
    :eitse_au_b => :d8,         # BS
    :eitse_au_e => :s6,         # BS
    :eitse_dt_b => :d9,         # TE, BS
    :eitse_dt_e => :s7,         # TE, BS
  }
  # 全局 actual 周期のみ
  All_A_C = Struct.new(:nit,    # TE, BS, CS110
                       :bit,    # TE, BS, CS110
                       :sdt,    # TE, BS, CS110
                       :eitpf,  # TE, BS, CS110
                       :tot,    # TE, BS, CS110
                       :sdtt,   # BS, CS110
                       :nbitr,  # CS110
                       )
  # 全局 actual 周期+範囲
  All_A_CR = Struct.new(:eitsb_tv_b,  # TE, BS, CS110
                        :eitsb_tv_e1, # TE, BS, CS110
                        :eitsb_tv_e2, # TE, BS, CS110
                        :eitsb_au_b,  # BS, CS110
                        :eitsb_au_e,  # BS, CS110
                        :eitsb_dt_b,  # TE, BS, CS110
                        :eitsb_dt_e,  # TE, BS, CS110
                        )
  # 全局 other 周期のみ
  All_O_C = Struct.new(:nit,         # CS110
                       :sdt,         # BS, CS110
                       :eitpf,       # BS, CS110
                       )
  # 全局 other 周期+範囲
  All_O_CR = Struct.new(:eitsb_tv_b,  # BS, CS110
                        :eitsb_tv_e1, # BS, CS110
                        :eitsb_tv_e2, # BS, CS110
                        :eitsb_au_b,  # BS, CS110
                        :eitsb_dt_b,  # BS, CS110
                        )
  # 全局SIプライムTS actual 周期のみ
  AllP_A_C = Struct.new(:sdt,
                        :nbitm,
                        :nbitr)
  # 全局SIプライムTS actual 周期+範囲
  AllP_A_CR = All_A_CR
  # 全局SIプライムTS other 周期のみ
  AllP_O_C = Struct.new(:sdt)
  # 全局SIプライムTS other 周期+範囲
  AllP_O_CR = All_A_CR
  
  # 各局 周期のみ
  Each_C = Struct.new(:sdtt,    # TE
                      :cdt,     # TE
                      :ldt,     # CS110
                      )
  # 各局 周期+範囲
  Each_CR = Struct.new(:eitsb_tv_b, # TE, BS, CS110
                       :eitsb_au_b, # BS, CS110
                       :eitsb_dt_b, # TE, BS, CS110
                       :eitse_tv_b, # TE, BS, CS110
                       :eitse_tv_e, # TE, BS, CS110
                       :eitse_au_b, # BS, CS110
                       :eitse_au_e, # BS, CS110
                       :eitse_dt_b, # TE, BS, CS110
                       :eitse_dt_e, # TE, BS, CS110
                       )
  # 範囲
  All_R = Struct.new(:d1, :d2, :d3, :s1, :s2, :s3, :s4)
  Each_R = Struct.new(:d4, :d5, :d6, :d7, :d8, :d9, :s5, :s6, :s7)
  # データ全体
  DataAll = Struct.new(:allar, :allac, :allacr, :alloc, :allocr)
  DataAllP = Struct.new(:allpac, :allpacr, :allpoc, :allpocr)
  DataEach = Struct.new(:eachr, :eachc, :eachcr)
  Type = Struct.new(:te, :bs, :cs110)
  def initialize
    @version = Type.new
    @versionp = Type.new
    @versione = Type.new({}, {}, {})
    @all = Type.new
    @allp = Type.new
    @each = Type.new({}, {}, {})
    default
  end
  def default
    @default_all = Type.new
    @default_allp = Type.new
    @default_each = Type.new
    @default_all.te = DataAll.new(All_R.new(8, nil, 2, 3, 13, nil, 0),
                                  All_A_C.new(1, 1, 2, 1, 5, nil, nil),
                                  All_A_CR.new(60, 3, 10, nil, nil, 60, 3),
                                  nil, nil)
    @default_each.te = DataEach.new(Each_R.new(15, nil, 8, 15, nil, 8, 3, nil, 0),
                                    Each_C.new(180, 600, nil),
                                    Each_CR.new(60, nil, 60, 60, 20, nil, nil, 60, 20))
    @default_all.bs = DataAll.new(All_R.new(8, 8, 2, 3, 13, 8, 0),
                                  All_A_C.new(10, 10, 3, 3, 5, 600, nil),
                                  All_A_CR.new(180, 10, 10, 180, 20, 360, 20),
                                  All_O_C.new(nil, 10, 10),
                                  All_O_CR.new(180, 10, 20, 180, 360))
    @default_each.bs = DataEach.new(Each_R.new(15, 15, 8, 15, 15, 8, 3, 3, 0),
                                    nil,
                                    Each_CR.new(180, 180, 360, 180, 20, 180, 20, 360, 20))
    @default_all.cs110 = DataAll.new(All_R.new(8, 8, 2, 3, 13, 8, 0),
                                     All_A_C.new(1, 10, 3, 3, 5, 600, 2),
                                     All_A_CR.new(180, 5, 15, 180, 20, 360, 20),
                                     All_O_C.new(10, 10, 3),
                                     All_O_CR.new(180, 5, 15, 180, 360))
    @default_allp.cs110 = DataAllP.new(AllP_A_C.new(5, 10, 10),
                                       AllP_A_CR.new(20, 20, 20, 20, 20, 20, 20),
                                       AllP_O_C.new(5),
                                       AllP_O_CR.new(20, 20, 20, 20, 20, 20, 20))
    @default_each.cs110 = DataEach.new(Each_R.new(15, 15, 8, 15, 15, 8, 3, 3, 0),
                                       Each_C.new(nil, nil, 180),
                                       Each_CR.new(180, 180, 360, 180, 20, 180, 20, 360, 20))
  end
  def analyze(type, bit)
    case type
    when :te
      analyze_all(:te, bit.descriptor)
      analyze_each(:te, bit.original_network_id, bit.data[0].descriptor)
    when :bs
      analyze_all(:bs, bit.descriptor)
      bit.each do |data|
        analyze_each(:bs, data.broadcaster_id, data.descriptor)
      end
    when :cs110
      analyze_all(:cs110, bit.descriptor)
      analyze_allp(:cs110, bit.descriptor)
      bit.each do |data|
        analyze_each(:cs110, data.broadcaster_id, data.descriptor)
      end
    end
  end
  def analyze_each(type, key, descriptor)
    descriptor.each do |desc|
      if desc.class == Descriptor::SIParameter
        next if desc.update_time > Date.today
        return if @versione[type][key] && @versione[type][key] == desc.update_time
        t = desc.table
        if t.eitsb
          eitsb_c = [ t.eitsb.tv.base_cycle,
                      t.eitsb.au.base_cycle,
                      t.eitsb.dt.base_cycle ]
          eitsb_r = [ t.eitsb.tv.schedule_range,
                      t.eitsb.au.schedule_range,
                      t.eitsb.dt.schedule_range]
        else
          eitsb_c = [ nil, nil, nil ]
        end
        if t.eitse
          eitse_c = [ t.eitse.tv.base_cycle,
                      t.eitse.tv.cycle_group[0].cycle,
                      t.eitse.au.base_cycle,
                      t.eitse.au.cycle_group[0].cycle,
                      t.eitse.dt.base_cycle,
                      t.eitse.dt.cycle_group[0].cycle ]
          eitse_r = [ t.eitse.tv.schedule_range,
                      t.eitse.au.schedule_range,
                      t.eitse.dt.schedule_range,
                      t.eitse.tv.cycle_group[0].num_of_segment,
                      t.eitse.au.cycle_group[0].num_of_segment,
                      t.eitse.dt.cycle_group[0].num_of_segment ]
        else
          eitse_c = [ nil, nil, nil, nil, nil, nil ]
        end
        @each[type][key] = DataEach.new(Each_R.new(*eitsb_r, *eitse_r),
                                        Each_C.new(t.sdtt, t.cdt, t.ldt),
                                        Each_CR.new(*eitsb_c,
                                                    *eitse_c))
        @versione[type][key] = desc.update_time
        p @each
        p @versione
      end
    end
  end
  def analyze_all(type, descriptor)
    descriptor.each do |desc|
      if desc.class == Descriptor::SIParameter
        next if desc.update_time > Date.today
        return if @version[type] && @version[type] == desc.update_time
        t = desc.table
        if type == :te
          eipf = t.eitpf.table_cycle_h
        else
          eipf = t.eitpf
        end
        if t.eitsb_o
          allocr = All_O_CR.new(t.eitsb_o.tv.base_cycle,
                                t.eitsb_o.tv.cycle_group[0].cycle,
                                t.eitsb_o.tv.cycle_group[1].cycle,
                                t.eitsb_o.au.base_cycle,
                                t.eitsb_o.dt.base_cycle)
        else
          allocr = nil
        end
        if t.eitsb
          allr = All_R.new(t.eitsb.tv.schedule_range,
                           t.eitsb.au.schedule_range,
                           t.eitsb.dt.schedule_range,
                           t.eitsb.tv.cycle_group[0].num_of_segment,
                           t.eitsb.tv.cycle_group[1].num_of_segment,
                           t.eitsb.au.cycle_group[0].num_of_segment,
                           t.eitsb.dt.cycle_group[0].num_of_segment)
          allacr = All_A_CR.new(t.eitsb.tv.base_cycle,
                                t.eitsb.tv.cycle_group[0].cycle,
                                t.eitsb.tv.cycle_group[1].cycle,
                                t.eitsb.au.base_cycle,
                                t.eitsb.au.cycle_group[0].cycle,
                                t.eitsb.dt.base_cycle,
                                t.eitsb.dt.cycle_group[0].cycle)
        else
          allr = nil
          allacr = nil
        end
        @all[type] = DataAll.new(allr,
                                 All_A_C.new(t.nit, t.bit, t.sdt, eipf, 5, t.sdtt, t.nbitr),
                                 allacr,
                                 All_O_C.new(t.nit_o, t.sdt_o, t.eitpf_o),
                                 allocr)
        @version[type] = desc.update_time
        p @all
        p @version
        break
      end
    end
  end
  def analyze_allp(type, descriptor)
    descriptor.each do |desc|
      if desc.class == Descriptor::SIPrimeTS
        next if desc.update_time > Date.today
        return if @version[type] && @version[type] == desc.update_time
        t = desc.table
        if t.eitsb
          allpacr = AllP_A_CR.new(t.eitsb.tv.base_cycle,
                                  t.eitsb.tv.cycle_group[0].cycle,
                                  t.eitsb.tv.cycle_group[1].cycle,
                                  t.eitsb.au.base_cycle,
                                  t.eitsb.au.cycle_group[0].cycle,
                                  t.eitsb.dt.base_cycle,
                                  t.eitsb.dt.cycle_group[0].cycle)
        else
          allpacr = nil
        end
        if t.eitsb_o
          allpocr = AllP_O_CR.new(t.eitsb_o.tv.base_cycle,
                                  t.eitsb_o.tv.cycle_group[0].cycle,
                                  t.eitsb_o.tv.cycle_group[1].cycle,
                                  t.eitsb_o.au.base_cycle,
                                  t.eitsb_o.au.cycle_group[0].cycle,
                                  t.eitsb_o.dt.base_cycle,
                                  t.eitsb_o.dt.cycle_group[0].cycle)
        else
          allpocr = nil
        end
        @allp[type] = DataAllP.new(AllP_A_C.new(t.sdt, t.nbitm, t.nbitr),
                                   allpacr,
                                   AllP_O_C.new(t.sdt_o),
                                   allpocr)
        @versionp[type] = desc.update_time
        p @allp
        p @versionp
        break
      end
    end
  end
end
