#include "ruby.h"

#include "isdb_decode_text.h"
#include "psisi.h"
#include "pes.h"
#include "mpegts.h"

#define PID_EMPTY	(0xffff)
#define PID_PAT		(0x0000)
#define PID_MAX		(0x1fff)

typedef struct {
	unsigned char sync_byte;
	unsigned char transport_error_indicator;
	unsigned char payload_unit_start_indicator;
	unsigned char transport_priority;
	unsigned int pid;
	unsigned char transport_scrambling_control;
	unsigned char adaptation_field_control;
	unsigned char continuity_counter;
} MpegTsHeader;

typedef struct {
	unsigned int pid;
	unsigned char counter;
	Psisidata *psisi;
} MpegTsSection;

typedef struct {
	/* セクション */
	int sec_size;
	MpegTsSection *sec;

	/* PES */
	int pes_flag;

	enum { EMPTY = 0, SECTION, PES } pid_tbl[PID_MAX];
} MpegTs;


HMpegTs mpegts_open(VALUE sec_pid, VALUE pes_flag)
{
	MpegTs *mpegts;
	long offset;

	mpegts = ALLOC(MpegTs);
	if (mpegts == NULL) {
		return NULL;
	}
	memset(mpegts, 0, sizeof(MpegTs));

	if (TYPE(sec_pid) == T_ARRAY) {
		mpegts->sec_size = RARRAY_LEN(sec_pid);
	}
	else {
		mpegts->sec_size = FIX2INT(sec_pid);
	}
	if (mpegts->sec_size < 0) {
		free(mpegts);
		return NULL;
	}
	if (mpegts->sec_size > 0) {
		mpegts->sec = ALLOC_N(MpegTsSection, mpegts->sec_size);
		if (mpegts->sec == NULL) {
			free(mpegts);
			return NULL;
		}
	}

	for (offset = 0; offset < mpegts->sec_size; offset++) {
		if (TYPE(sec_pid) == T_ARRAY) {
			mpegts->sec[offset].pid = FIX2INT(rb_ary_entry(sec_pid, offset));
			if (mpegts->sec[offset].pid > PID_MAX) {
				free(mpegts->sec);
				free(mpegts);
				return NULL;
			}
			mpegts->pid_tbl[mpegts->sec[offset].pid] = SECTION;
		}
		else
			mpegts->sec[offset].pid = PID_EMPTY;
		mpegts->sec[offset].counter = 0xff;
		mpegts->sec[offset].psisi = NULL;
	}

	mpegts->pes_flag = RTEST(pes_flag);

	return (HMpegTs)mpegts;
}

void mpegts_close(HMpegTs hmpegts)
{
	MpegTs *mpegts;
	int i;

	mpegts = (MpegTs *)hmpegts;

	if (mpegts != NULL) {
		for (i = 0; i < mpegts->sec_size; i++) {
			if (mpegts->sec[i].psisi != NULL) {
				psisi_free(mpegts->sec[i].psisi);
			}
		}
		if (mpegts->sec != NULL) {
			free(mpegts->sec);
		}

		free(mpegts);
	}
}

static VALUE mpegts_section_start(MpegTsSection *mpegtssec,
								  unsigned char **packet, int *len)
{
	VALUE psisi;
	int n;

	if (mpegtssec->psisi != NULL)
		psisi_free(mpegtssec->psisi);
	mpegtssec->psisi = psisi_create(*packet, *len, mpegtssec->pid);
	if (mpegtssec->psisi == NULL)
		return Qnil;

	n = psisi_add(mpegtssec->psisi, *packet, *len);
	*packet += n;
	*len -= n;

	psisi = psisi_make(&mpegtssec->psisi);
	if (psisi == Qnil) {
		return Qnil;
	}
	else if (psisi == Qfalse) {
		return Qnil;
	}
	else {
		return psisi;
	}
}

static VALUE mpegts_section_add(MpegTsSection *mpegtssec,
								unsigned char **packet, int *len)
{
	VALUE psisi;
	int n;

	if (mpegtssec->psisi == NULL) {
		return Qfalse;
	}

	n = psisi_add(mpegtssec->psisi, *packet, *len);
	*packet += n;
	*len -= n;

	psisi = psisi_make(&mpegtssec->psisi);
	if (psisi == Qnil) {
		return Qnil;
	}
	else if (psisi == Qfalse) {
		return Qnil;
	}
	else {
		return psisi;
	}
}

static int mpegts_counter_check(MpegTsSection *mpegtssec,
								unsigned char continuity_counter)
{
	if (mpegtssec->counter == 0xff) {
		mpegtssec->counter = continuity_counter;
		return 0;
	}
	else {
		mpegtssec->counter++;
		if (mpegtssec->counter == 0x10)
			mpegtssec->counter = 0;
		if (mpegtssec->counter != continuity_counter) {
			mpegtssec->counter = continuity_counter;
			return 1;
		}
		return 0;
	}
}

static void mpegts_section_init(MpegTsSection *mpegtssec)
{
	if (mpegtssec->psisi != NULL) {
		psisi_free(mpegtssec->psisi);
		mpegtssec->psisi = NULL;
	}
}

VALUE mpegtssec_analyze(MpegTsHeader *hdr, MpegTsSection *mpegtssec,
						unsigned char *packet, int len)
{
	VALUE rslt;
	VALUE psisi;
	int ptr;
	Psisidata *psisidata;

	if (mpegts_counter_check(mpegtssec, hdr->continuity_counter)) {
		mpegts_section_init(mpegtssec);
	}

	rslt = rb_ary_new2(10);
	if (hdr->payload_unit_start_indicator == 1) {
		if (packet[0] == 0) {
			packet++;
			len--;
			psisi = mpegts_section_start(mpegtssec, &packet, &len);
		}
		else {
			ptr = packet[0];
			packet++;
			len--;
			psisi = mpegts_section_add(mpegtssec, &packet, &len);
			if (psisi == Qfalse) {
				packet += ptr;
				len -= len;
				psisi = mpegts_section_start(mpegtssec, &packet, &len);
			}
		}
	}
	else {
		psisi = mpegts_section_add(mpegtssec, &packet, &len);
		if (psisi == Qfalse) {
			psisi = Qnil;
		}
	}
	if (!NIL_P(psisi)) {
		GetPsisi(psisi, psisidata);
		if (psisidata->section_syntax_indicator == 1)
			rb_ary_push(rslt, psisi);
	}

	while (len > 0 && *packet != 0xff) {
		psisi = mpegts_section_start(mpegtssec, &packet, &len);
		if (!NIL_P(psisi)) {
			GetPsisi(psisi, psisidata);
			if (psisidata->section_syntax_indicator == 1)
				rb_ary_push(rslt, psisi);
		}
	}

	return rslt;
}

VALUE mpegtspes_analyze(MpegTsHeader *hdr,
						unsigned char *packet, int len)
{
	if (hdr->payload_unit_start_indicator == 1) {
		return peshdr_create(packet, len, hdr->pid);
	}
	else
		return pesbdy_create(packet, len, hdr->pid);
}

VALUE mpegts_analyze(HMpegTs hmpegts, unsigned char *packet, int len)
{
	MpegTs *mpegts;
	MpegTsHeader hdr;
	MpegTsSection *mpegtssec;
	unsigned char adaptation_field_length;
	int i;
	VALUE rslt;

	mpegts = (MpegTs *)hmpegts;
	if (mpegts == NULL) {
		return Qnil;
	}

	hdr.sync_byte = packet[0];
	hdr.transport_error_indicator = (packet[1] >> 7) & 0x1;
	hdr.payload_unit_start_indicator = (packet[1] >> 6) & 0x1;
	hdr.transport_priority = (packet[1] >> 5) & 0x1;
	hdr.pid = ((packet[1] & 0x1f) << 8) | packet[2];
	hdr.transport_scrambling_control = (packet[3] >> 6) & 0x3;
	hdr.adaptation_field_control = (packet[3] >> 4) & 0x3;
	hdr.continuity_counter = packet[3] & 0xf;

	if ((hdr.adaptation_field_control & 1) == 0)
		return Qnil;
	packet += 4;
	len -= 4;
	if (hdr.adaptation_field_control & 2) {
		adaptation_field_length = *packet++;
		len--;
		packet += adaptation_field_length;
		len -= adaptation_field_length;
	}

	if (mpegts->pid_tbl[hdr.pid] == SECTION) {
		/* セクション: 登録済みPIDを処理する */
		mpegtssec = NULL;
		for (i = 0; i < mpegts->sec_size; i++) {
			if (mpegts->sec[i].pid == hdr.pid) {
				mpegtssec = &mpegts->sec[i];
				break;
			}
			else if (mpegts->sec[i].pid == PID_EMPTY)
				break;
		}
		if (mpegtssec)
			return mpegtssec_analyze(&hdr, mpegtssec, packet, len);
	}
	else if (mpegts->pid_tbl[hdr.pid] == PES) {
		/* PES: 登録済みPIDを処理する */
		if (mpegts->pes_flag)
			return mpegtspes_analyze(&hdr, packet, len);
	}
	else if (mpegts->pid_tbl[hdr.pid] == EMPTY) {
		/* 未登録PIDを処理する */
		if (hdr.payload_unit_start_indicator == 1) {
			if (hdr.pid == PID_PAT)
				rslt = Qnil;
			else
				rslt = peshdr_create(packet, len, hdr.pid);
			if (rslt == Qnil) {
				/* セクションだった場合 */
				mpegts->pid_tbl[hdr.pid] = SECTION;
				mpegtssec = NULL;
				for (i = 0; i < mpegts->sec_size; i++) {
					if (mpegts->sec[i].pid == PID_EMPTY) {
						mpegts->sec[i].pid = hdr.pid;
						mpegtssec = &mpegts->sec[i];
						break;
					}
				}
				if (mpegtssec)
					return mpegtssec_analyze(&hdr, mpegtssec, packet, len);
			}
			else {
				/* PESだった場合 */
				mpegts->pid_tbl[hdr.pid] = PES;
				if (mpegts->pes_flag)
					return rslt;
			}
		}
	}

	return Qnil;
}

/*
 * Local Variables:
 * c-basic-offset: 4
 * indent-tabs-mode: t
 * tab-width: 4
 * End:
 */
