#include "ruby.h"
#include "ruby/encoding.h"

#include <stdio.h>

#include "isdb_decode_text.h"
#include "crc32.h"
#include "psisi.h"

/* memo:8単位文字符号の2バイト文字をUTF-8に変換すると最大4バイトになる */
#define ISDB_BUF_MAX	(SECTION_MAX * 2)

VALUE rb_cPsisi, rb_ePsisiError;
VALUE rb_cPsisi_PAT_Data;
VALUE rb_cPsisi_PMT_Data;

static void free_psisi(Psisidata *psisi)
{
	if (psisi) {
		if (psisi->isdb_buf)
			free(psisi->isdb_buf);
		if (psisi->isdb_decode)
			isdb_decode_close(psisi->isdb_decode);
		free(psisi);
	}
}

static VALUE psisi_alloc(VALUE klass)
{
	return Data_Wrap_Struct(klass, 0, free_psisi, 0);
}

static VALUE psisi_body(VALUE self)
{
	Psisidata *psisi;

	GetPsisi(self, psisi);
	if (psisi->section_syntax_indicator == 1)
		return rb_str_new((const char *)psisi->section.buf + 8,
						  psisi->section.size - 8 - 4);
	else
		return rb_str_new((const char *)psisi->section.buf + 3,
						  psisi->section.size - 3);
}

static VALUE psisi_inspect(VALUE self)
{
	Psisidata *psisi;
	char str[128];

	GetPsisi(self, psisi);
	sprintf(str, "<Psisi table_id:0x%02x pid:0x%04x>",
			psisi->table_id, psisi->pid);
	return rb_str_new(str, strlen(str));
}

static VALUE psisi_text(VALUE self, VALUE str)
{
	Psisidata *psisi;
	const unsigned char *src;
	unsigned int src_len;
	unsigned int len;

	Check_Type(str, T_STRING);
	GetPsisi(self, psisi);
	if (psisi->isdb_decode == NULL)
		psisi->isdb_decode = isdb_decode_open(ISDB_ARIB);
	if (psisi->isdb_buf == NULL)
		psisi->isdb_buf = ALLOC_N(unsigned char, ISDB_BUF_MAX);
	if (psisi->isdb_decode == NULL || psisi->isdb_buf == NULL)
		return Qnil;

	src = (const unsigned char *)RSTRING_PTR(str);
	src_len = (unsigned int)RSTRING_LEN(str);
	len = isdb_decode_text(psisi->isdb_decode, src, src_len,
						   psisi->isdb_buf, ISDB_BUF_MAX);

	return rb_enc_str_new((const char *)psisi->isdb_buf, (long)len,
						  rb_utf8_encoding());
}

#define PSISI_DEFFUNC1(xxxx)					\
	static VALUE psisi_##xxxx(VALUE self)		\
{												\
	Psisidata *psisi;							\
												\
	GetPsisi(self, psisi);						\
	return INT2FIX(psisi->xxxx);				\
}
#define PSISI_DEFFUNC2(xxxx)					\
	static VALUE psisi_##xxxx(VALUE self)		\
{												\
	Psisidata *psisi;							\
												\
	GetPsisi(self, psisi);						\
	if (psisi->section_syntax_indicator == 0)	\
		return Qnil;							\
	return INT2FIX(psisi->xxxx);				\
}
#define PMT_DEFFUNC(xxxx)						\
	static VALUE pmt_##xxxx(VALUE self)			\
{												\
	Psisidata *psisi;							\
												\
	GetPsisi(self, psisi);						\
	if (psisi->table_id != PMT)					\
		return Qnil;							\
	return INT2FIX(psisi->u.pmt.xxxx);			\
}
PSISI_DEFFUNC1(table_id)
PSISI_DEFFUNC1(section_syntax_indicator)
PSISI_DEFFUNC1(section_length)
PSISI_DEFFUNC2(table_id_extention)
PSISI_DEFFUNC2(version_number)
PSISI_DEFFUNC2(current_next_indicator)
PSISI_DEFFUNC2(section_number)
PSISI_DEFFUNC2(last_section_number)
PSISI_DEFFUNC1(pid)
PMT_DEFFUNC(pcr_pid)

static VALUE pat_data(VALUE self)
{
	Psisidata *psisi;
	VALUE rslt;
	VALUE data;
	int i;

	GetPsisi(self, psisi);
	if (psisi->table_id != PAT)
		return Qnil;

	rslt = rb_ary_new2(psisi->u.pat.data_size);
	for (i = 0; i < psisi->u.pat.data_size; i++) {
		data = rb_ary_new3(2, INT2FIX(psisi->u.pat.data[i].program_number),
						   INT2FIX(psisi->u.pat.data[i].pid));
		rb_ary_push(rslt, data);
	}

	return rslt;
}

static VALUE pmt_program_info(VALUE self)
{
	Psisidata *psisi;

	GetPsisi(self, psisi);
	if (psisi->table_id != PMT)
		return Qnil;

	return rb_str_new((const char *)psisi->u.pmt.program_info,
					  psisi->u.pmt.program_info_length);
}

static VALUE pmt_data(VALUE self)
{
	Psisidata *psisi;
	VALUE rslt;
	VALUE data;
	VALUE es_info;
	int i;

	GetPsisi(self, psisi);
	if (psisi->table_id != PMT)
		return Qnil;

	rslt = rb_ary_new2(psisi->u.pmt.data_size);
	for (i = 0; i < psisi->u.pmt.data_size; i++) {
		es_info = rb_str_new((const char *)psisi->u.pmt.data[i].es_info,
							 psisi->u.pmt.data[i].es_info_length);
		data = rb_ary_new3(3, INT2FIX(psisi->u.pmt.data[i].stream_type),
						   INT2FIX(psisi->u.pmt.data[i].elementary_pid),
						   es_info);
		rb_ary_push(rslt, data);
	}

	return rslt;
}

#define GETBIT16(p)	(((p)[0] << 8) | (p)[1])
#define GETBIT13(p)	((((p)[0] & 0x1f) << 8) | (p)[1])
#define GETBIT12(p)	((((p)[0] & 0xf) << 8) | (p)[1])
Psisidata *psisi_create(unsigned char *sec, int len, unsigned int pid)
{
	unsigned char section_syntax_indicator;
	unsigned int section_length;
	Psisidata *psisi;

	if (sec == NULL || len < 3)
		return NULL;
	section_syntax_indicator = (sec[1] >> 7) & 1;
	section_length = GETBIT12(sec + 1);
	if ((section_syntax_indicator == 1 && len < 8) ||
		sizeof(psisi->section.buf) < section_length)
		return NULL;

	psisi = ALLOC(Psisidata);
	if (psisi == NULL)
		return NULL;

	psisi->pid = pid;
	psisi->section.size = section_length + 3;
	psisi->section.current_size = 0;
	psisi->isdb_decode = NULL;
	psisi->isdb_buf = NULL;

	psisi->table_id = sec[0];
	psisi->section_syntax_indicator = section_syntax_indicator;
	psisi->section_length = section_length;
	if (section_syntax_indicator == 1) {
		psisi->table_id_extention = GETBIT16(sec + 3);
		psisi->version_number = (sec[5] >> 1) & 0x1f;
		psisi->current_next_indicator = sec[5] & 1;
		psisi->section_number = sec[6];
		psisi->last_section_number = sec[7];
	}

	return psisi;
}

void psisi_free(Psisidata *psisi)
{
	free_psisi(psisi);
}

int psisi_add(Psisidata *psisi, unsigned char *sec, int len)
{
	if (psisi == NULL || sec == NULL || len < 0)
		return 0;

	if (psisi->section.size < (psisi->section.current_size + len)) {
		len = psisi->section.size - psisi->section.current_size;
	}
	if (sizeof(psisi->section.buf) < (psisi->section.current_size + len)) {
		len = sizeof(psisi->section.buf) - psisi->section.current_size;
	}
	if (len < 0)
		return 0;

	memcpy(psisi->section.buf + psisi->section.current_size, sec, len);
	psisi->section.current_size += len;

	return len;
}

VALUE psisi_make(Psisidata **psisi_p)
{
	Psisidata *psisi;
	VALUE self;
	unsigned char *p, *p_end;
	int n;

	if (psisi_p == NULL || *psisi_p == NULL)
		return Qnil;

	psisi = *psisi_p;
	if (psisi->section.size > psisi->section.current_size)
		return Qnil;

	if (psisi->section_syntax_indicator == 1) {
		if (crc(psisi->section.buf, psisi->section.size) != 0) {
			psisi_free(psisi);
			*psisi_p = NULL;
			return Qfalse;
		}

		p = psisi->section.buf + 8;
		p_end = psisi->section.buf + psisi->section.size - 4;
	}
	else {
		p = psisi->section.buf + 3;
		p_end = psisi->section.buf + psisi->section.size;
	}

	switch (psisi->table_id) {
	case PAT:
		n = 0;
		while (p < p_end) {
			psisi->u.pat.data[n].program_number = GETBIT16(p);
			p += 2;
			psisi->u.pat.data[n].pid = GETBIT13(p);
			p += 2;
			n++;
			if (n >= PAT_DATA_MAX)
				break;
		}
		psisi->u.pat.data_size = n;
		break;

	case PMT:
		psisi->u.pmt.pcr_pid = GETBIT13(p);
		p += 2;
		psisi->u.pmt.program_info_length = GETBIT12(p);
		p += 2;
		psisi->u.pmt.program_info = p;
		p += psisi->u.pmt.program_info_length;

		n = 0;
		while (p < p_end) {
			psisi->u.pmt.data[n].stream_type = p[0];
			p += 1;
			psisi->u.pmt.data[n].elementary_pid = GETBIT13(p);
			p += 2;
			psisi->u.pmt.data[n].es_info_length = GETBIT12(p);
			p += 2;
			psisi->u.pmt.data[n].es_info = p;
			p += psisi->u.pmt.data[n].es_info_length;
			n++;
			if (n >= PMT_DATA_MAX)
				break;
		}
		psisi->u.pmt.data_size = n;
		break;

	default:
		break;
	}

	self = rb_funcall(rb_cPsisi, rb_intern("new"), 0);
	DATA_PTR(self) = psisi;
	*psisi_p = NULL;
	return self;
}

void Init_psisi(void)
{
    rb_cPsisi = rb_define_class("Psisi", rb_cObject);
    rb_ePsisiError = rb_define_class("PsisiError", rb_eStandardError);

    rb_define_alloc_func(rb_cPsisi, psisi_alloc);

	rb_define_method(rb_cPsisi, "table_id", psisi_table_id, 0);
	rb_define_method(rb_cPsisi, "section_syntax_indicator", psisi_section_syntax_indicator, 0);
	rb_define_method(rb_cPsisi, "section_length", psisi_section_length, 0);
	rb_define_method(rb_cPsisi, "table_id_extention", psisi_table_id_extention, 0);
	rb_define_method(rb_cPsisi, "version_number", psisi_version_number, 0);
	rb_define_method(rb_cPsisi, "current_next_indicator", psisi_current_next_indicator, 0);
	rb_define_method(rb_cPsisi, "section_number", psisi_section_number, 0);
	rb_define_method(rb_cPsisi, "last_section_number", psisi_last_section_number, 0);

	rb_define_method(rb_cPsisi, "pid", psisi_pid, 0);
	rb_define_method(rb_cPsisi, "body", psisi_body, 0);
	rb_define_method(rb_cPsisi, "text", psisi_text, 1);
	rb_define_method(rb_cPsisi, "inspect", psisi_inspect, 0);

	/* PAT */
	rb_const_set(rb_cPsisi, rb_intern("PAT"), INT2FIX(PAT));
	rb_cPsisi_PAT_Data = rb_funcall(rb_cStruct, rb_intern("new"), 2,
									ID2SYM(rb_intern("program_number")),
									ID2SYM(rb_intern("pid")));
	rb_const_set(rb_cPsisi, rb_intern("PAT_Data"), rb_cPsisi_PAT_Data);
	rb_define_method(rb_cPsisi, "pat_data", pat_data, 0);

	/* PMT */
	rb_const_set(rb_cPsisi, rb_intern("PMT"), INT2FIX(PMT));
	rb_cPsisi_PMT_Data = rb_funcall(rb_cStruct, rb_intern("new"), 3,
									ID2SYM(rb_intern("stream_type")),
									ID2SYM(rb_intern("elementary_pid")),
									ID2SYM(rb_intern("descriptor")));
	rb_const_set(rb_cPsisi, rb_intern("PMT_Data"), rb_cPsisi_PMT_Data);
	rb_define_method(rb_cPsisi, "pmt_pcr_pid", pmt_pcr_pid, 0);
	rb_define_method(rb_cPsisi, "pmt_program_info", pmt_program_info, 0);
	rb_define_method(rb_cPsisi, "pmt_data", pmt_data, 0);
}

/*
 * Local Variables:
 * c-basic-offset: 4
 * indent-tabs-mode: t
 * tab-width: 4
 * End:
 */
