#include "ruby.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <limits.h>
#include <linux/dvb/frontend.h>
#include <linux/dvb/dmx.h>

#include "isdb_decode_text.h"

#include "v4ldvb.h"
#include "mpegts.h"
#include "psisi.h"
#include "pes.h"

typedef struct {
	int fd;
} DvbFrontendData;

typedef struct {
	int fd;
} DvbDemuxData;

typedef struct {
	int fd;
	struct sigaction backup_sa;

	unsigned char *packet;
	HMpegTs hmpegts;
} DvbDvrData;

#define GetDvbFrontend(obj, dvbfrontend) {\
	Data_Get_Struct((obj), DvbFrontendData, (dvbfrontend));\
	if ((dvbfrontend) == 0) rb_raise(rb_eDvbFrontendError, "closed DvbFrontend");\
}

#define GetDvbDemux(obj, dvbdemux) {\
	Data_Get_Struct((obj), DvbDemuxData, (dvbdemux));\
	if ((dvbdemux) == 0) rb_raise(rb_eDvbDemuxError, "closed DvbDemux");\
}

#define GetDvbDvr(obj, dvbdvr) {\
	Data_Get_Struct((obj), DvbDvrData, (dvbdvr));\
	if ((dvbdvr) == 0) rb_raise(rb_eDvbDvrError, "closed DvbDvr");\
}

VALUE rb_cDvbAdapter;
VALUE rb_cDvbFrontend, rb_eDvbFrontendError;
VALUE rb_cDvbDemux, rb_eDvbDemuxError;
VALUE rb_cDvbDvr, rb_eDvbDvrError;

#define SYNC_BYTE	(0x47)

static VALUE dvbadapter_initialize(VALUE self, VALUE n, VALUE m, VALUE type)
{
	rb_ivar_set(self, rb_intern("@n"), n);
	rb_ivar_set(self, rb_intern("@m"), m);
	rb_ivar_set(self, rb_intern("@type"), type);

	return self;
}

static void free_dvbfrontend(DvbFrontendData *dvbfrontend)
{
	if (dvbfrontend) {
		if (dvbfrontend->fd >= 0)
			close(dvbfrontend->fd);
		free(dvbfrontend);
	}
}

static VALUE dvbfrontend_alloc(VALUE klass)
{
	return Data_Wrap_Struct(klass, 0, free_dvbfrontend, 0);
}

static VALUE dvbfrontend_initialize(VALUE self)
{
	DvbFrontendData *dvbfrontend;

	dvbfrontend = ALLOC(DvbFrontendData);
	DATA_PTR(self) = dvbfrontend;
	dvbfrontend->fd = (-1);

	return self;
}

static void free_dvbdemux(DvbDemuxData *dvbdemux)
{
	if (dvbdemux) {
		if (dvbdemux->fd >= 0)
			close(dvbdemux->fd);
		free(dvbdemux);
	}
}

static VALUE dvbdemux_alloc(VALUE klass)
{
	return Data_Wrap_Struct(klass, 0, free_dvbdemux, 0);
}

static VALUE dvbdemux_initialize(VALUE self)
{
	DvbDemuxData *dvbdemux;

	dvbdemux = ALLOC(DvbDemuxData);
	DATA_PTR(self) = dvbdemux;
	dvbdemux->fd = (-1);

	return self;
}

static void free_dvbdvr(DvbDvrData *dvbdvr)
{
	if (dvbdvr) {
		if (dvbdvr->fd >= 0)
			close(dvbdvr->fd);

		if (dvbdvr->packet)
			free(dvbdvr->packet);
		if (dvbdvr->hmpegts)
			mpegts_close(dvbdvr->hmpegts);
		free(dvbdvr);
	}
}

static VALUE dvbdvr_alloc(VALUE klass)
{
	return Data_Wrap_Struct(klass, 0, free_dvbdvr, 0);
}

static VALUE dvbdvr_initialize(VALUE self)
{
	DvbDvrData *dvbdvr;

	dvbdvr = ALLOC(DvbDvrData);
	DATA_PTR(self) = dvbdvr;
	dvbdvr->fd = (-1);
	dvbdvr->packet = NULL;
	dvbdvr->hmpegts = NULL;

	return self;
}

static void dvbadapter_getname(VALUE self, char *filename, const char *str)
{
	sprintf(filename, "/dev/dvb/adapter%d/%s%d",
			FIX2INT(rb_ivar_get(self, rb_intern("@n"))),
			str,
			FIX2INT(rb_ivar_get(self, rb_intern("@m"))));
}

static void dvbfrontend_getname(VALUE self, char *filename, const char *str)
{
	dvbadapter_getname(rb_ivar_get(self, rb_intern("@adapter")),
					   filename, str);
}

static void dvbdemux_getname(VALUE self, char *filename, const char *str)
{
	dvbfrontend_getname(rb_ivar_get(self, rb_intern("@dvbfrontend")),
						filename, str);
}

static void dvbdvr_getname(VALUE self, char *filename, const char *str)
{
	dvbfrontend_getname(rb_ivar_get(self, rb_intern("@dvbfrontend")),
						filename, str);
}

static VALUE dvbfrontend_gettype(VALUE self)
{
	return rb_funcall(rb_ivar_get(self, rb_intern("@adapter")),
					  rb_intern("type"), 0);
}

static VALUE dvbdemux_gettype(VALUE self)
{
	return dvbfrontend_gettype(rb_ivar_get(self, rb_intern("@dvbfrontend")));
}

typedef enum { TE, BS, CS110 } DeliveryType;
static DeliveryType dvbfrontend_getCtype(VALUE self)
{
	VALUE type;

	type = dvbfrontend_gettype(self);
	if (type == ID2SYM(rb_intern("bs")))
		return BS;
	else if (type == ID2SYM(rb_intern("cs110")))
		return CS110;
	else
		return TE;
}

static DeliveryType dvbdemux_getCtype(VALUE self)
{
	return dvbfrontend_getCtype(rb_ivar_get(self, rb_intern("@dvbfrontend")));
}

static void dvbfrontend_open(VALUE self)
{
	DvbFrontendData *dvbfrontend;
	char filename[PATH_MAX+1];
	int fd;

	GetDvbFrontend(self, dvbfrontend);
	dvbfrontend_getname(self, filename, "frontend");
	fd = open(filename, O_RDWR);
	if (fd < 0) {
		rb_raise(rb_eDvbFrontendError, "frontend open error");
		return;
	}

	if (dvbfrontend->fd >= 0)
		close(dvbfrontend->fd);
	dvbfrontend->fd = fd;
}

static void dvbfrontend_close(VALUE self)
{
	DvbFrontendData *dvbfrontend;
	int fd;

	GetDvbFrontend(self, dvbfrontend);
	fd = dvbfrontend->fd;
	dvbfrontend->fd = (-1);

	close(fd);
}

static VALUE dvbfrontend_setvoltage(VALUE self, VALUE lnb)
{
	DvbFrontendData *dvbfrontend;
	fe_sec_voltage_t voltage;

	GetDvbFrontend(self, dvbfrontend);

	voltage = SEC_VOLTAGE_OFF;
	if (lnb == ID2SYM(rb_intern("voltage13"))) {
		voltage = SEC_VOLTAGE_13;
	}
	else if (lnb == ID2SYM(rb_intern("voltage18"))) {
		voltage = SEC_VOLTAGE_18;
	}
	else if (lnb == ID2SYM(rb_intern("voltageoff"))) {
		voltage = SEC_VOLTAGE_OFF;
	}
#if 1
	struct dtv_properties param;
	struct dtv_property props[1];

	memset(&param, 0, sizeof(param));
	memset(&props, 0, sizeof(props));
	param.props = props;

	param.props[param.num].cmd = DTV_VOLTAGE;
	param.props[param.num].u.data = voltage;
	param.num++;

	if (ioctl(dvbfrontend->fd, FE_SET_PROPERTY, &param) < 0) {
		rb_raise(rb_eDvbFrontendError, "frontend FE_SET_PROPERTY error");
		return Qnil;
	}
#else
	if (ioctl(dvbfrontend->fd, FE_SET_VOLTAGE, &voltage) < 0) {
		rb_raise(rb_eDvbFrontendError, "frontend FE_SET_VOLTAGE error");
		return Qnil;
	}
#endif

	return Qtrue;
}

static VALUE dvbfrontend_setoneshot(VALUE self)
{
	DvbFrontendData *dvbfrontend;
	unsigned int flags;

	GetDvbFrontend(self, dvbfrontend);

	flags = FE_TUNE_MODE_ONESHOT;
	if (ioctl(dvbfrontend->fd, FE_SET_FRONTEND_TUNE_MODE, flags) < 0) {
		rb_raise(rb_eDvbFrontendError, "frontend FE_SET_FRONTEND_TUNE_MODE error");
		return Qfalse;
	}

	return Qtrue;
}

static VALUE dvbfrontend_ch2frequency(VALUE self, VALUE ch)
{
	int freq;
	DeliveryType type;

	type = dvbfrontend_getCtype(self);

	if (type == BS) {
		freq = 11727480 + (38360 / 2) * (FIX2INT(ch) - 1);
	}
	else if (type == CS110) {
		freq = 12291000 + (40000 / 2) * (FIX2INT(ch) - 2);
	}
	else {
		freq = 473142857 + 6000000 * (FIX2INT(ch) - 13);
	}

	return INT2FIX(freq);
}

static VALUE dvbfrontend_tune(int argc, VALUE *argv, VALUE self)
{
	DvbFrontendData *dvbfrontend;
	VALUE ch;
	VALUE tsid;
	int freq;
#if 1
	struct dtv_properties param;
	struct dtv_property props[3];
#else
    struct dvb_frontend_parameters param;
#endif
	DeliveryType type;
	fe_status_t status;
	int i;

	GetDvbFrontend(self, dvbfrontend);
	rb_scan_args(argc, argv, "11", &ch, &tsid);
	if (argc < 2)
		tsid = Qnil;

	type = dvbfrontend_getCtype(self);
	freq = FIX2INT(dvbfrontend_ch2frequency(self, ch));

#if 1
	memset(&param, 0, sizeof(param));
	memset(&props, 0, sizeof(props));
	param.props = props;

	if (type == BS || type == CS110) {
		param.props[param.num].cmd = DTV_FREQUENCY;
		param.props[param.num].u.data = freq - 10678000;
		param.num++;
	}
	else {
		param.props[param.num].cmd = DTV_FREQUENCY;
		param.props[param.num].u.data = freq;
		param.num++;
	}
	if (tsid != Qnil) {
		param.props[param.num].cmd = DTV_ISDBS_TS_ID;
		param.props[param.num].u.data = FIX2INT(tsid);
		param.num++;
	}
	param.props[param.num].cmd = DTV_TUNE;
	param.num++;

	if (ioctl(dvbfrontend->fd, FE_SET_PROPERTY, &param) < 0) {
		rb_raise(rb_eDvbFrontendError, "frontend FE_SET_PROPERTY error");
		return Qnil;
	}
#else
	memset(&param, 0, sizeof(param));
	if (type == BS || type == CS110) {
		param.frequency = freq - 10678000;
		param.inversion = INVERSION_AUTO;
		param.u.qpsk.symbol_rate = 28860000;
		param.u.qpsk.fec_inner = FEC_AUTO;
	}
	else {
		param.frequency = freq;
		param.inversion = INVERSION_AUTO;
		param.u.ofdm.bandwidth = BANDWIDTH_AUTO;
		param.u.ofdm.code_rate_HP = FEC_AUTO;
		param.u.ofdm.code_rate_LP = FEC_AUTO;
		param.u.ofdm.constellation = QAM_AUTO;
		param.u.ofdm.transmission_mode = TRANSMISSION_MODE_AUTO;
		param.u.ofdm.guard_interval = GUARD_INTERVAL_AUTO;
		param.u.ofdm.hierarchy_information = HIERARCHY_AUTO;
	}

	if (ioctl(dvbfrontend->fd, FE_SET_FRONTEND, &param) < 0) {
		rb_raise(rb_eDvbFrontendError, "frontend FE_SET_FRONTEND error");
		return Qnil;
	}
#endif

	status = 0;
	for (i = 0; i < 10; i++) {
		if (ioctl(dvbfrontend->fd, FE_READ_STATUS, &status) < 0) {
			rb_raise(rb_eDvbFrontendError, "frontend FE_READ_STATUS error");
			return Qnil;
		}

		if (status & FE_HAS_LOCK)
			break;

		usleep(50 * 1000);
	}
	if (status & FE_HAS_LOCK) {
		return INT2FIX(freq);
	}
	else {
		return Qnil;
	}

}

static void dvbdemux_open(VALUE self)
{
	DvbDemuxData *dvbdemux;
	char filename[PATH_MAX+1];
	int fd;

	GetDvbDemux(self, dvbdemux);
	dvbdemux_getname(self, filename, "demux");
	fd = open(filename, O_RDWR|O_NONBLOCK);
	if (fd < 0) {
		rb_raise(rb_eDvbDemuxError, "demux open error");
		return;
	}

	if (dvbdemux->fd >= 0)
		close(dvbdemux->fd);
	dvbdemux->fd = fd;
}

static void dvbdemux_close(VALUE self)
{
	DvbDemuxData *dvbdemux;
	int fd;

	GetDvbDemux(self, dvbdemux);
	fd = dvbdemux->fd;
	dvbdemux->fd = (-1);

	close(fd);
}

static VALUE dvbdemux_setsctfilter(int argc, VALUE *argv, VALUE self)
{
	DvbDemuxData *dvbdemux;
	struct dmx_sct_filter_params sct;
	VALUE pid;
	VALUE filters;
	VALUE filter;
	int i;

	GetDvbDemux(self, dvbdemux);
	rb_scan_args(argc, argv, "1*", &pid, &filters);

	memset(&sct, 0, sizeof(sct));
	sct.pid = FIX2INT(pid);
	for (i = 0; i < RARRAY_LEN(filters); i++) {
		filter = rb_ary_entry(filters, i);
		sct.filter.filter[i] = FIX2INT(filter);
		sct.filter.mask[i] = 0xff;
	}
	sct.flags = DMX_IMMEDIATE_START;

	if (ioctl(dvbdemux->fd, DMX_SET_FILTER, &sct) < 0) {
		rb_raise(rb_eDvbDemuxError, "demux DMX_SET_FILTER error");
		return Qnil;
	}

	return Qtrue;
}

static VALUE dvbdemux_setpesfilter(int argc, VALUE *argv, VALUE self)
{
	DvbDemuxData *dvbdemux;
	struct dmx_pes_filter_params pes;
	VALUE pid;
	VALUE pestype;
	dmx_pes_type_t pestype_int;

	GetDvbDemux(self, dvbdemux);
	rb_scan_args(argc, argv, "02", &pid, &pestype);
	if (argc < 1)
		pid = INT2FIX(0x2000);
	if (argc < 2) {
		pestype_int = DMX_PES_OTHER;
	}
	else {
		if (pestype == ID2SYM(rb_intern("video")))
			pestype_int = DMX_PES_VIDEO;
		else if (pestype == ID2SYM(rb_intern("audio")))
			pestype_int = DMX_PES_AUDIO;
		else if (pestype == ID2SYM(rb_intern("subtitle")))
			pestype_int = DMX_PES_SUBTITLE;
		else if (pestype == ID2SYM(rb_intern("teletext")))
			pestype_int = DMX_PES_TELETEXT;
		else if (pestype == ID2SYM(rb_intern("pcr")))
			pestype_int = DMX_PES_PCR;
		else if (pestype == ID2SYM(rb_intern("other")))
			pestype_int = DMX_PES_OTHER;
		else
			pestype_int = FIX2INT(pestype);
	}

	memset(&pes, 0, sizeof(pes));
	pes.pid = FIX2INT(pid);
	pes.input = DMX_IN_FRONTEND;
	pes.output = DMX_OUT_TS_TAP;
	pes.pes_type = pestype_int;
	pes.flags = DMX_IMMEDIATE_START;
	if (ioctl(dvbdemux->fd, DMX_SET_PES_FILTER, &pes) < 0) {
		rb_raise(rb_eDvbDemuxError, "demux DMX_SET_PES_FILTER error");
		return Qnil;
	}

	return Qtrue;
}

static VALUE dvbdemux_start(VALUE self)
{
	DvbDemuxData *dvbdemux;

	GetDvbDemux(self, dvbdemux);
	if (ioctl(dvbdemux->fd, DMX_START) < 0)
		return Qfalse;

	return Qtrue;
}

static VALUE dvbdemux_stop(VALUE self)
{
	DvbDemuxData *dvbdemux;

	GetDvbDemux(self, dvbdemux);
	if (ioctl(dvbdemux->fd, DMX_STOP) < 0)
		return Qfalse;

	return Qtrue;
}

static void dvbdvr_open(VALUE self)
{
	DvbDvrData *dvbdvr;
	char filename[PATH_MAX+1];
	int fd;

	GetDvbDvr(self, dvbdvr);
	dvbdvr_getname(self, filename, "dvr");
	fd = open(filename, O_RDONLY|O_NONBLOCK);
	if (fd < 0) {
		rb_raise(rb_eDvbDvrError, "dvr open error");
		return;
	}

	if (dvbdvr->fd >= 0)
		close(dvbdvr->fd);
	dvbdvr->fd = fd;
}

static void dvbdvr_close(VALUE self)
{
	DvbDvrData *dvbdvr;
	int fd;

	GetDvbDvr(self, dvbdvr);
	fd = dvbdvr->fd;
	dvbdvr->fd = (-1);

	close(fd);
}

static VALUE dvbfrontend_s_close(VALUE self)
{
	dvbfrontend_close(self);

	return Qnil;
}

static VALUE dvbfrontend_s_open(int argc, VALUE *argv, VALUE klass)
{
	VALUE self;
	VALUE adapter;

	if (!rb_block_given_p()) {
		rb_raise(rb_eArgError, "block is not given");
		return Qnil;
	}
	rb_scan_args(argc, argv, "1", &adapter);

	self = rb_funcall(klass, rb_intern("new"), 0);
	rb_ivar_set(self, rb_intern("@adapter"), adapter);

	dvbfrontend_open(self);

	return rb_ensure(rb_yield, self, dvbfrontend_s_close, self);
}

static VALUE dvbdemux_s_close(VALUE self)
{
	int i;

	if (TYPE(self) == T_ARRAY) {
		for (i = 0; i < RARRAY_LEN(self); i++) {
			dvbdemux_close(rb_ary_entry(self, i));
		}
	}
	else {
		dvbdemux_close(self);
	}

	return Qnil;
}

static VALUE dvbdemux_s_open(int argc, VALUE *argv, VALUE klass)
{
	VALUE ary;
	VALUE self;
	VALUE dvbfrontend;
	VALUE n;
	int i;

	if (!rb_block_given_p()) {
		rb_raise(rb_eArgError, "block is not given");
		return Qnil;
	}
	rb_scan_args(argc, argv, "11", &dvbfrontend, &n);

	if (n == Qnil) {
		self = rb_funcall(klass, rb_intern("new"), 0);
		rb_ivar_set(self, rb_intern("@dvbfrontend"), dvbfrontend);

		dvbdemux_open(self);

		return rb_ensure(rb_yield, self, dvbdemux_s_close, self);
	}
	else {
		ary = rb_ary_new2(FIX2INT(n));
		for (i = 0; i < FIX2INT(n); i++) {
			self = rb_funcall(klass, rb_intern("new"), 0);
			rb_ivar_set(self, rb_intern("@dvbfrontend"), dvbfrontend);

			dvbdemux_open(self);
			rb_ary_push(ary, self);
		}

		return rb_ensure(rb_yield, ary, dvbdemux_s_close, ary);
	}
}

static VALUE dvbdvr_s_close(VALUE self)
{
	DvbDvrData *dvbdvr;

	GetDvbDvr(self, dvbdvr);
	sigaction(SIGPIPE, &dvbdvr->backup_sa, NULL);
	dvbdvr_close(self);

	return Qnil;
}

static VALUE dvbdvr_s_open(int argc, VALUE *argv, VALUE klass)
{
	DvbDvrData *dvbdvr;
	VALUE self;
	VALUE frontend;
	struct sigaction sa;

	if (!rb_block_given_p()) {
		rb_raise(rb_eArgError, "block is not given");
		return Qnil;
	}
	rb_scan_args(argc, argv, "1", &frontend);

	self = rb_funcall(klass, rb_intern("new"), 0);
	GetDvbDvr(self, dvbdvr);
	rb_ivar_set(self, rb_intern("@dvbfrontend"), frontend);
	rb_ivar_set(self, rb_intern("@ts_packet_size"), INT2FIX(188));
	rb_ivar_set(self, rb_intern("@ts_read_size"), INT2FIX(4096));

	dvbdvr_open(self);

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sigaction(SIGPIPE, &sa, &dvbdvr->backup_sa);

	return rb_ensure(rb_yield, self, dvbdvr_s_close, self);
}

static int dvbdemux_read(int fd, int timeout, VALUE *psisi)
{
	fd_set fds;
	struct timeval tv;
	int len, seclen;
	unsigned char buff[4096];
	int i;
	Psisidata *psisidata;

	for (i = 0; i < 3; i++) {
		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		memset(&tv, 0, sizeof(tv));
		tv.tv_sec = timeout / 1000;
		tv.tv_usec = (timeout % 1000) * 1000;
		if (rb_thread_select(fd+1, &fds, NULL, NULL, &tv) > 0) {
#if 1
			len = read(fd, buff, sizeof(buff));
			psisidata = psisi_create(buff, len, 0xffff);
			if (psisidata == NULL) {
				*psisi = Qnil;
				break;
			}
			seclen = psisidata->section.size;
			if (len != seclen)
				rb_raise(rb_eDvbDemuxError, "len:%d, secstion.size:%d \n", len, seclen);
			psisi_add(psisidata, buff, seclen);
			*psisi = psisi_make(&psisidata);
			if (*psisi == Qnil)
				;
			else if (*psisi == Qfalse)
				;
			else
				break;
			if (psisidata != NULL)
				psisi_free(psisidata);
#else
			len = read(fd, buff, 8);
			if (len == 8) {
				psisidata = psisi_create(buff, len, 0xffff);
				if (psisidata == NULL) {
					*psisi = Qnil;
					break;
				}
				seclen = psisidata->section.size;
				len = read(fd, buff + 8, seclen - 8) + 8;
				if (len > 8) {
					psisi_add(psisidata, buff, seclen);
					*psisi = psisi_make(&psisidata);
					if (*psisi == Qnil)
						;
					else if (*psisi == Qfalse)
						;
					else
						break;
				}
				if (psisidata != NULL)
					psisi_free(psisidata);
			}
#endif
		}
		else
			return 0;
	}

	return 1;
}

static VALUE dvbdemux_section(VALUE self, VALUE timeout)
{
	DvbDemuxData *dvbdemux;
	VALUE rslt;
	VALUE psisi;

	GetDvbDemux(self, dvbdemux);
	rslt = Qnil;

	for (;;) {
		psisi = Qnil;
		if (dvbdemux_read(dvbdemux->fd, FIX2INT(timeout), &psisi) == 0) {
			/* time out */
			break;
		}

		if (psisi != Qnil)
			rslt = rb_yield(psisi);
	}

	return rslt;
}

static VALUE dvbdemux_scan(VALUE self, VALUE timeout)
{
	DvbDemuxData *dvbdemux;
	VALUE dvbfrontend;
	VALUE rslt;
	DeliveryType type;
	int ch_start, ch_end, ch_step;
	int ch;
	VALUE freq;
	VALUE tsids;
	VALUE argv[2];
	VALUE psisi;
	int i;

	GetDvbDemux(self, dvbdemux);
	dvbfrontend = rb_ivar_get(self, rb_intern("@dvbfrontend"));
	rslt = Qnil;

	dvbdemux_stop(self);

	type = dvbdemux_getCtype(self);
	if (type == BS) {
		ch_start = 1;
		ch_end = 23;
		ch_step = 2;
	}
	else if (type == CS110) {
		ch_start = 2;
		ch_end = 24;
		ch_step = 2;
	}
	else {
		ch_start = 13;
		ch_end = 62;
		ch_step = 1;
	}

	ch = ch_start;
	while (ch <= ch_end) {
		tsids = rb_yield(rb_ary_new3(2, ID2SYM(rb_intern("gettsid")),
									 dvbfrontend_ch2frequency(dvbfrontend,
															  INT2FIX(ch))));
		i = 0;
		do {
			if (tsids == Qnil) {
				argv[0] = INT2FIX(ch);
				freq = dvbfrontend_tune(1, argv, dvbfrontend);
			}
			else {
				argv[0] = INT2FIX(ch);
				argv[1] = rb_ary_entry(tsids, i);
				freq = dvbfrontend_tune(2, argv, dvbfrontend);
			}
			rb_yield(rb_ary_new3(2, ID2SYM(rb_intern("tune")),
								 rb_ary_new3(2, INT2FIX(ch), freq)));

			dvbdemux_start(self);

			psisi = Qnil;
			dvbdemux_read(dvbdemux->fd, FIX2INT(timeout), &psisi);
			if (psisi != Qnil)
				rslt = rb_yield(rb_ary_new3(2, ID2SYM(rb_intern("found")),
													psisi));
			else
				rb_yield(ID2SYM(rb_intern("notfound")));

			dvbdemux_stop(self);

			i++;
		}
		while (tsids != Qnil && i < RARRAY_LEN(tsids));

		ch += ch_step;
	}

	return rslt;
}

static VALUE dvbdvr_setfilter(VALUE self, VALUE sec_pid, VALUE pes_flag)
{
	DvbDvrData *dvbdvr;

	GetDvbDvr(self, dvbdvr);

	if (dvbdvr->hmpegts != NULL) {
		mpegts_close(dvbdvr->hmpegts);
	}
	dvbdvr->hmpegts = mpegts_open(sec_pid, pes_flag);

	return Qtrue;
}

static int dvbdvr_select(int fd, int timeout)
{
	fd_set fds;
	struct timeval tv;

	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	memset(&tv, 0, sizeof(tv));
	tv.tv_sec = timeout / 1000;
	tv.tv_usec = (timeout % 1000) * 1000;
	if (rb_thread_select(fd+1, &fds, NULL, NULL, &tv) > 0) {
		return 1;
	}
	else
		return 0;
}

static VALUE dvbdvr_section(VALUE self, VALUE timeout, VALUE socket)
{
	DvbDvrData *dvbdvr;
	VALUE rslt;
	int ts_packet_size;
	int ts_read_size;
	int sd;
	int len, prev_len;
	unsigned char *p;
	int i;
	VALUE psisi_pes, psisi;

	GetDvbDvr(self, dvbdvr);
	ts_packet_size = FIX2INT(rb_ivar_get(self, rb_intern("@ts_packet_size")));
	ts_read_size = FIX2INT(rb_ivar_get(self, rb_intern("@ts_read_size")));
	if (socket == Qnil)
		sd = (-1);
	else
		sd = FIX2INT(rb_funcall(socket, rb_intern("fileno"), 0, 0));
	rslt = Qnil;

	if (dvbdvr->packet == NULL)	/* 一度だけ確保 */
		dvbdvr->packet = ALLOC_N(unsigned char, ts_read_size);

	prev_len = 0;
	for (;;) {
	loop:
		if (dvbdvr_select(dvbdvr->fd, FIX2INT(timeout)) == 0) {
			/* time out */
			break;
		}
		len = read(dvbdvr->fd, dvbdvr->packet + prev_len,
				   ts_read_size - prev_len) + prev_len;

		if (len < ts_packet_size) {
			prev_len = len;
			goto loop;
		}
		prev_len = 0;
		if (!rb_block_given_p() || dvbdvr->hmpegts == NULL) {
			/* ブロックなし or setfilterなしのときはyieldしない */
			if (sd >= 0) {
				if (write(sd, dvbdvr->packet, len) < 0) {
					/* SIGPIPE etc. */
					break;
				}
			}
			goto loop;
		}

		for (p = dvbdvr->packet;
			 len - (p - dvbdvr->packet) >= ts_packet_size;
			 p += ts_packet_size) {
			if (p[0] != SYNC_BYTE) {
				if (sd >= 0) {
					if (write(sd, dvbdvr->packet, p - dvbdvr->packet) < 0) {
						/* SIGPIPE etc. */
						break;
					}
				}

				/* SYNC_BYTEを探す */
				p = memchr(p, SYNC_BYTE, len - (p - dvbdvr->packet));
				if (p != NULL) {
					/* 見つかった場合はそこからを次回に繰越 */
					prev_len = len - (p - dvbdvr->packet);
					memmove(dvbdvr->packet, p, prev_len);
					goto loop;
				}
				goto loop;
			}
			psisi_pes = mpegts_analyze(dvbdvr->hmpegts, p, ts_packet_size);
			if (TYPE(psisi_pes) == T_ARRAY) {
				for (i = 0; i < RARRAY_LEN(psisi_pes); i++) {
					psisi = rb_ary_entry(psisi_pes, i);
					rslt = rb_yield(psisi);
				}
			}
			else if (psisi_pes == Qnil) {
			}
			else {
				rslt = rb_yield(psisi_pes);
			}
		}

		/* 半端分を次回に繰越 */
		prev_len = len - (p - dvbdvr->packet);
		if (sd >= 0) {
			if (write(sd, dvbdvr->packet, len - prev_len) < 0) {
				/* SIGPIPE etc. */
				break;
			}
		}
		if (prev_len > 0) {
			memmove(dvbdvr->packet, p, prev_len);
		}
	}

	return rslt;
}

static VALUE dvbfrontend_getinfo(VALUE self)
{
	DvbFrontendData *dvbfrontend;
	struct dvb_frontend_info info;
	VALUE rslt;
	VALUE symbol;

	GetDvbFrontend(self, dvbfrontend);

	memset(&info, 0, sizeof(info));

	if (ioctl(dvbfrontend->fd, FE_GET_INFO, &info) < 0) {
		rb_raise(rb_eDvbFrontendError, "frontend FE_GET_INFO error");
		return Qfalse;
	}

	rslt = rb_hash_new();
	rb_hash_aset(rslt, ID2SYM(rb_intern("name")),
				 rb_str_new2(info.name));

	symbol = Qnil;
	switch (info.type) {
	case FE_QPSK:				/* DVB-S */
		symbol = ID2SYM(rb_intern("qpsk"));
		break;
	case FE_QAM:				/* DVB-C */
		symbol = ID2SYM(rb_intern("qam"));
		break;
	case FE_OFDM:				/* DVB-T */
		symbol = ID2SYM(rb_intern("ofdm"));
		break;
	default:
		break;
	}
	rb_hash_aset(rslt, ID2SYM(rb_intern("type")), symbol);

	rb_hash_aset(rslt, ID2SYM(rb_intern("frequency_min")),
				 INT2FIX(info.frequency_min));
	rb_hash_aset(rslt, ID2SYM(rb_intern("frequency_max")),
				 INT2FIX(info.frequency_max));
	rb_hash_aset(rslt, ID2SYM(rb_intern("frequency_stepsize")),
				 INT2FIX(info.frequency_stepsize));
	rb_hash_aset(rslt, ID2SYM(rb_intern("frequency_tolerance")),
				 INT2FIX(info.frequency_tolerance));
	rb_hash_aset(rslt, ID2SYM(rb_intern("symbol_rate_min")),
				 INT2FIX(info.symbol_rate_min));
	rb_hash_aset(rslt, ID2SYM(rb_intern("symbol_rate_max")),
				 INT2FIX(info.symbol_rate_max));
	rb_hash_aset(rslt, ID2SYM(rb_intern("symbol_rate_tolerance")),
				 INT2FIX(info.symbol_rate_tolerance));
	rb_hash_aset(rslt, ID2SYM(rb_intern("notifier_delay")),
				 INT2FIX(info.notifier_delay));
	rb_hash_aset(rslt, ID2SYM(rb_intern("caps")),
				 INT2FIX(info.caps));

	return rslt;
}

static VALUE dvbfrontend_getdelivery(VALUE self)
{
	DvbFrontendData *dvbfrontend;
	struct dtv_properties param;
	struct dtv_property props[1];
	VALUE rslt;

	GetDvbFrontend(self, dvbfrontend);

	memset(&param, 0, sizeof(param));
	memset(&props, 0, sizeof(props));
	param.props = props;

	param.props[param.num].cmd = DTV_DELIVERY_SYSTEM;
	param.num++;

	if (ioctl(dvbfrontend->fd, FE_GET_PROPERTY, &param) < 0) {
		rb_raise(rb_eDvbFrontendError, "frontend FE_GET_PROPERTY error");
		return Qfalse;
	}

	rslt = Qnil;
	switch (param.props[0].u.data) {
	case SYS_UNDEFINED:	rslt = ID2SYM(rb_intern("undefined"));	break;
	case SYS_DVBC_ANNEX_AC:	rslt = ID2SYM(rb_intern("dvbc_annex_ac"));	break;
	case SYS_DVBC_ANNEX_B:	rslt = ID2SYM(rb_intern("dvbc_annex_b"));	break;
	case SYS_DVBT:	rslt = ID2SYM(rb_intern("dvbt"));	break;
	case SYS_DSS:	rslt = ID2SYM(rb_intern("dss"));	break;
	case SYS_DVBS:	rslt = ID2SYM(rb_intern("dvbs"));	break;
	case SYS_DVBS2:	rslt = ID2SYM(rb_intern("dvbs2"));	break;
	case SYS_DVBH:	rslt = ID2SYM(rb_intern("dvbh"));	break;
	case SYS_ISDBT:	rslt = ID2SYM(rb_intern("isdbt"));	break;
	case SYS_ISDBS:	rslt = ID2SYM(rb_intern("isdbs"));	break;
	case SYS_ISDBC:	rslt = ID2SYM(rb_intern("isdbc"));	break;
	case SYS_ATSC:	rslt = ID2SYM(rb_intern("atsc"));	break;
	case SYS_ATSCMH:	rslt = ID2SYM(rb_intern("atscmh"));	break;
	case SYS_DMBTH:	rslt = ID2SYM(rb_intern("dmbth"));	break;
	case SYS_CMMB:	rslt = ID2SYM(rb_intern("cmmb"));	break;
	case SYS_DAB:	rslt = ID2SYM(rb_intern("dab"));	break;
	default:
		break;
	}

	return rslt;
}

void Init_v4ldvb(void)
{
	/* DvbAdapter class */
    rb_cDvbAdapter = rb_define_class("DvbAdapter", rb_cObject);
	rb_define_method(rb_cDvbAdapter, "initialize", dvbadapter_initialize, 3);
	rb_define_attr(rb_cDvbAdapter, "n", 1, 0);
	rb_define_attr(rb_cDvbAdapter, "m", 1, 0);
	rb_define_attr(rb_cDvbAdapter, "type", 1, 0);

	/* DvbFrontend class */
    rb_cDvbFrontend = rb_define_class("DvbFrontend", rb_cObject);
    rb_eDvbFrontendError = rb_define_class("DvbFrontendError", rb_eStandardError);

    rb_define_alloc_func(rb_cDvbFrontend, dvbfrontend_alloc);

	rb_define_singleton_method(rb_cDvbFrontend, "open", dvbfrontend_s_open, -1);
	rb_define_method(rb_cDvbFrontend, "initialize", dvbfrontend_initialize, 0);
	rb_define_method(rb_cDvbFrontend, "type", dvbfrontend_gettype, 0);
	rb_define_method(rb_cDvbFrontend, "setvoltage", dvbfrontend_setvoltage, 1);
	rb_define_method(rb_cDvbFrontend, "setoneshot", dvbfrontend_setoneshot, 0);
	rb_define_method(rb_cDvbFrontend, "ch2frequency", dvbfrontend_ch2frequency, 1);
	rb_define_method(rb_cDvbFrontend, "tune", dvbfrontend_tune, -1);
	rb_define_method(rb_cDvbFrontend, "getinfo", dvbfrontend_getinfo, 0);
	rb_define_method(rb_cDvbFrontend, "getdelivery", dvbfrontend_getdelivery, 0);

	/* DvbDemux class */
    rb_cDvbDemux = rb_define_class("DvbDemux", rb_cObject);
    rb_eDvbDemuxError = rb_define_class("DvbDemuxError", rb_eStandardError);

    rb_define_alloc_func(rb_cDvbDemux, dvbdemux_alloc);

	rb_define_singleton_method(rb_cDvbDemux, "open", dvbdemux_s_open, -1);
	rb_define_method(rb_cDvbDemux, "initialize", dvbdemux_initialize, 0);
	rb_define_method(rb_cDvbDemux, "type", dvbdemux_gettype, 0);
	rb_define_method(rb_cDvbDemux, "setsctfilter", dvbdemux_setsctfilter, -1);
	rb_define_method(rb_cDvbDemux, "setpesfilter", dvbdemux_setpesfilter, -1);
	rb_define_method(rb_cDvbDemux, "start", dvbdemux_start, 0);
	rb_define_method(rb_cDvbDemux, "stop", dvbdemux_stop, 0);
	rb_define_method(rb_cDvbDemux, "section", dvbdemux_section, 1);
	rb_define_method(rb_cDvbDemux, "scan", dvbdemux_scan, 1);

	/* DvbDvr class */
    rb_cDvbDvr = rb_define_class("DvbDvr", rb_cObject);
    rb_eDvbDvrError = rb_define_class("DvbDvrError", rb_eStandardError);

    rb_define_alloc_func(rb_cDvbDvr, dvbdvr_alloc);

	rb_define_singleton_method(rb_cDvbDvr, "open", dvbdvr_s_open, -1);
	rb_define_method(rb_cDvbDvr, "initialize", dvbdvr_initialize, 0);
	rb_define_method(rb_cDvbDvr, "setfilter", dvbdvr_setfilter, 2);
	rb_define_method(rb_cDvbDvr, "section", dvbdvr_section, 2);

	Init_psisi();
	Init_pes();
}

/*
 * Local Variables:
 * c-basic-offset: 4
 * indent-tabs-mode: t
 * tab-width: 4
 * End:
 */
